from pydicom.dataset import Dataset
from pydicom.uid import generate_uid
import SimpleITK as sitk
import numpy as np
from pydicomrt.reg.builder import DeformableSpatialRegistrationBuilder

# -----------------------------------------------------------
# Build fake data
# -----------------------------------------------------------
def make_fake_fixed_ref_ds() -> Dataset:
    ds = Dataset()
    ds.PatientName = "Test^Patient"
    ds.PatientID = "TP001"
    ds.PatientBirthDate = "19700101"
    ds.PatientSex = "O"
    ds.StudyInstanceUID = generate_uid()
    ds.StudyID = "1"
    ds.StudyDate = "20250101"
    ds.StudyTime = "120000"
    ds.AccessionNumber = ""
    ds.ReferringPhysicianName = ""
    ds.FrameOfReferenceUID = generate_uid()
    ds.PositionReferenceIndicator = "SCI"
    return ds

def make_fake_moving_instance(frame_of_ref_uid: str) -> Dataset:
    ds = Dataset()
    ds.SOPClassUID = "1.2.840.10008.5.1.4.1.1.2"  # CT Image Storage
    ds.SOPInstanceUID = generate_uid()
    ds.FrameOfReferenceUID = frame_of_ref_uid
    return ds

def make_displacement_field_transform(
    size=(8, 9, 10),
    spacing=(1.5, 1.5, 2.0),
    origin=(0.0, 0.0, 0.0),
    direction=(1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0),
    shift=(1.0, 0.0, 0.0)
) -> sitk.DisplacementFieldTransform:
    img_x = sitk.Image(size, sitk.sitkFloat64)
    img_x += float(shift[0])
    img_y = sitk.Image(size, sitk.sitkFloat64)
    img_y += float(shift[1])
    img_z = sitk.Image(size, sitk.sitkFloat64)
    img_z += float(shift[2])

    vec = sitk.Compose(img_x, img_y, img_z)
    vec = sitk.Cast(vec, sitk.sitkVectorFloat64)

    vec.SetSpacing(spacing)
    vec.SetOrigin(origin)
    vec.SetDirection(direction)

    tfm = sitk.DisplacementFieldTransform(vec)

    tfm.GetOrigin = lambda: vec.GetOrigin()
    tfm.GetSpacing = lambda: vec.GetSpacing()
    tfm.GetSize = lambda: vec.GetSize()
    tfm.GetDirection = lambda: vec.GetDirection()

    return tfm

def identity_4x4_flat_f32():
    M = np.eye(4, dtype=np.float32)
    return [float(x) for x in M.reshape(-1)]

# -----------------------------------------------------------
# Demo
# -----------------------------------------------------------
def demo():
    fixed_ref = make_fake_fixed_ref_ds()

    moving_list = [
        make_fake_moving_instance(fixed_ref.FrameOfReferenceUID),
        make_fake_moving_instance(fixed_ref.FrameOfReferenceUID),
    ]

    disp_tfm = make_displacement_field_transform(
        size=(8, 9, 10),
        spacing=(1.5, 1.5, 2.0),
        origin=(0.0, 0.0, 0.0),
        direction=(1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0),
        shift=(1.0, 0.0, 0.0)
    )
    preM = identity_4x4_flat_f32()
    postM = identity_4x4_flat_f32()

    # 4) Builder: Add a deformable registration block and build the REG dataset
    builder = DeformableSpatialRegistrationBuilder([fixed_ref])
    builder.set_uid_prefix("1.2.826.0.1.3680043.2.1125.")
    builder.add_deformable_registration(
        moving_ds_list=moving_list,
        vectorial_field_transform=disp_tfm,
        pre_transform=preM,
        post_transform=postM
    )
    reg_ds = builder.build()

    # 5) Verify the output
    print("=== REG Dataset Summary ===")
    print("SOPClassUID: ", reg_ds.SOPClassUID)
    print("SOPInstanceUID:", reg_ds.SOPInstanceUID)
    print("Modality   : ", reg_ds.Modality)
    print("SeriesDesc : ", getattr(reg_ds, "SeriesDescription", ""))
    print("FoR UID    : ", getattr(reg_ds, "FrameOfReferenceUID", ""))

    seq = reg_ds.DeformableRegistrationSequence
    print("\nDeformableRegistrationSequence items:", len(seq))
    item = seq[0]
    print("  SourceFoR UID:", item.SourceFrameOfReferenceUID)
    print("  #ReferencedImages:", len(item.ReferencedImageSequence))

    grid = item.DeformableRegistrationGridSequence[0]
    print("  GridDimensions : ", grid.GridDimensions)
    print("  GridResolution : ", grid.GridResolution)
    print("  IOP            : ", list(grid.ImageOrientationPatient))
    print("  IPP            : ", list(grid.ImagePositionPatient))
    print("  VectorGridData bytes:", len(grid.VectorGridData))
    code = item.RegistrationTypeCodeSequence[0]
    print("  Code (Value,Meaning):", code.CodeValue, code.CodeMeaning)

    # 6)（Optional）：
    # from pydicom.filewriter import write_file
    # write_file("REG_Deformable.dcm", reg_ds)
    # print("\nSaved to REG_Deformable.dcm")


if __name__ == "__main__":
    demo()



from pydicom.uid import UID

# RT Dose CIOD
RT_DOSE_IOD = {
    "PatientID": {},
    "PatientName": {},
    "PatientBirthDate": {},
    "PatientSex": {},
    "StudyInstanceUID": {},
    "StudyID": {},
    "StudyDate": {},
    "StudyTime": {},
    "AccessionNumber": {},
    "SeriesInstanceUID": {"type": UID},
    "SeriesNumber": {},
    "Modality": {"value": "RTDOSE"},
    "Manufacturer": {},
    "FrameOfReferenceUID": {"type": UID},
    "SOPClassUID": {"type": UID},
    "SOPInstanceUID": {"type": UID},
    "InstanceNumber": {},

    "NumberOfFrames": {},
    "Rows": {},
    "Columns": {},
    "PixelSpacing": {},
    "ImageOrientationPatient": {},
    "ImagePositionPatient": {},     # if value == (3004,000C) GridFrameOffsetVector is required
    "SliceThickness": {"optional": True},

    "BitsAllocated": {},
    "BitsStored": {},
    "HighBit": {},
    "PixelRepresentation": {},

    "DoseUnits": {"value": "GY"},
    "DoseType": {},
    "DoseSummationType": {"value": "PLAN"},
    "GridFrameOffsetVector": {},

    "PixelData": {},
}



from pydicom.dataset import Dataset
from .ds_reg_ds_iod import DEFORMABLE_SPATIAL_REGSITRATION_IOD
from .s_reg_ds_iod import SPATIAL_REGSITRATION_IOD
from pydicomrt.utils.validate_dcm_info import check_iod


def check_ds_reg_iod(rs_ds: Dataset) -> dict:
    result_dict = {
        "result": True,
        "content": []
    }
    failed_item_list = []
    failed_item_list = check_iod(ds=rs_ds, config_map=DEFORMABLE_SPATIAL_REGSITRATION_IOD)
    if len(failed_item_list) > 0:
        result_dict["result"] = False
        result_dict["content"] = failed_item_list
    return result_dict


def check_s_reg_iod(rs_ds: Dataset) -> dict:
    result_dict = {
        "result": True,
        "content": []
    }
    failed_item_list = []
    failed_item_list = check_iod(ds=rs_ds, config_map=SPATIAL_REGSITRATION_IOD)
    if len(failed_item_list) > 0:
        result_dict["result"] = False
        result_dict["content"] = failed_item_list
    return result_dict



import os
import datetime

from pydicom.uid import generate_uid
from pydicom.dataset import Dataset, FileDataset, FileMetaDataset
from pydicom.sequence import Sequence
from pydicom.uid import ImplicitVRLittleEndian


DICOM_UID_PREFIX = os.getenv('DICOM_UID_PREFIX', "1.2.826.0.1.3680043.8.498.")  # PYDICOM_ROOT_UID

def create_rtstruct_dataset(series_data) -> FileDataset:
    ds = generate_base_dataset()
    add_study_and_series_information(ds, series_data)
    add_patient_information(ds, series_data)
    add_refd_frame_of_ref_sequence(ds, series_data)
    add_rs_series_information(ds, series_data)
    return ds

def generate_base_dataset() -> FileDataset:
    file_name = "rs_test"
    file_meta = get_file_meta()
    ds = FileDataset(file_name, {}, file_meta=file_meta, preamble=b"\0" * 128)
    add_required_elements_to_ds(ds)
    add_sequence_lists_to_ds(ds)
    return ds

def get_file_meta() -> FileMetaDataset:
    file_meta = FileMetaDataset()
    file_meta.FileMetaInformationGroupLength = 202
    file_meta.FileMetaInformationVersion = b"\x00\x01"
    file_meta.TransferSyntaxUID = ImplicitVRLittleEndian
    file_meta.MediaStorageSOPClassUID = "1.2.840.10008.5.1.4.1.1.481.3"
    file_meta.MediaStorageSOPInstanceUID = (
        generate_uid(prefix=DICOM_UID_PREFIX)
    )
    file_meta.ImplementationClassUID = DICOM_UID_PREFIX + "1"
    # TODO file_meta.ImplementationVersionName = ""
    return file_meta

def add_required_elements_to_ds(ds: FileDataset):
    dt = datetime.datetime.now()
    # Append data elements required by the DICOM standarad
    ds.SpecificCharacterSet = "ISO_IR 192"
    ds.InstanceCreationDate = dt.strftime("%Y%m%d")
    ds.InstanceCreationTime = dt.strftime("%H%M%S")
    ds.StructureSetLabel = "RS_" + dt.strftime("%Y%m%d")
    ds.StructureSetDate = dt.strftime("%Y%m%d")
    ds.StructureSetTime = dt.strftime("%H%M%S.%f")
    ds.Modality = "RTSTRUCT"
    ds.Manufacturer = "higumalu"
    ds.ManufacturerModelName = "modelv1"
    ds.InstitutionName = "higumalu"
    # Set the transfer syntax
    ds.is_little_endian = True
    ds.is_implicit_VR = True
    # Set values already defined in the file meta
    ds.SOPClassUID = ds.file_meta.MediaStorageSOPClassUID
    ds.SOPInstanceUID = ds.file_meta.MediaStorageSOPInstanceUID

    ds.ApprovalStatus = "UNAPPROVED"

def add_sequence_lists_to_ds(ds: FileDataset):
    ds.StructureSetROISequence = Sequence()
    ds.ROIContourSequence = Sequence()
    ds.RTROIObservationsSequence = Sequence()

# ---------------------------------------------------------------------------------------- #

def add_study_and_series_information(ds: FileDataset, series_data):
    reference_ds = series_data[0]  # All elements in series should have the same data
    ds.StudyDate = reference_ds.StudyDate
    ds.SeriesDate = getattr(reference_ds, "SeriesDate", "")
    ds.StudyTime = reference_ds.StudyTime
    ds.SeriesTime = getattr(reference_ds, "SeriesTime", "")
    ds.StudyDescription = getattr(reference_ds, "StudyDescription", "")
    ds.SeriesDescription = getattr(reference_ds, "SeriesDescription", "")
    ds.StudyInstanceUID = reference_ds.StudyInstanceUID
    ds.SeriesInstanceUID = generate_uid(prefix=DICOM_UID_PREFIX)  # TODO: find out if random generation is ok
    ds.StudyID = reference_ds.StudyID
    ds.AccessionNumber = getattr(reference_ds, "AccessionNumber", "")
    ds.SeriesNumber = "1"  # TODO: find out if we can just use 1 (Should be fine since its a new series)
    ds.OperatorsName = getattr(reference_ds, "OperatorsName", "")

def add_patient_information(ds: FileDataset, series_data):
    reference_ds = series_data[0]  # All elements in series should have the same data
    ds.PatientName = getattr(reference_ds, "PatientName", "")
    ds.PatientID = getattr(reference_ds, "PatientID", "")
    ds.PatientBirthDate = getattr(reference_ds, "PatientBirthDate", "")
    ds.PatientSex = getattr(reference_ds, "PatientSex", "")
    ds.PatientAge = getattr(reference_ds, "PatientAge", "")
    ds.PatientSize = getattr(reference_ds, "PatientSize", "")
    ds.PatientWeight = getattr(reference_ds, "PatientWeight", "")

def add_refd_frame_of_ref_sequence(ds: FileDataset, series_data):
    refd_frame_of_ref = Dataset()
    refd_frame_of_ref.FrameOfReferenceUID = getattr(series_data[0], 'FrameOfReferenceUID', generate_uid(prefix=DICOM_UID_PREFIX))
    refd_frame_of_ref.RTReferencedStudySequence = create_frame_of_ref_study_sequence(series_data)

    ds.ReferencedFrameOfReferenceSequence = Sequence()
    ds.ReferencedFrameOfReferenceSequence.append(refd_frame_of_ref)
    ds.FrameOfReferenceUID = getattr(series_data[0], 'FrameOfReferenceUID', generate_uid(prefix=DICOM_UID_PREFIX))

def create_frame_of_ref_study_sequence(series_data) -> Sequence:
    reference_ds = series_data[0]
    rt_refd_series = Dataset()
    rt_refd_series.SeriesInstanceUID = reference_ds.SeriesInstanceUID
    rt_refd_series.ContourImageSequence = create_contour_image_sequence(series_data)
    rt_refd_series_sequence = Sequence()
    rt_refd_series_sequence.append(rt_refd_series)
    rt_refd_study = Dataset()
    rt_refd_study.ReferencedSOPClassUID = "1.2.840.10008.3.1.2.3.1"
    rt_refd_study.ReferencedSOPInstanceUID = reference_ds.StudyInstanceUID
    rt_refd_study.RTReferencedSeriesSequence = rt_refd_series_sequence
    rt_refd_study_sequence = Sequence()
    rt_refd_study_sequence.append(rt_refd_study)
    return rt_refd_study_sequence

def create_contour_image_sequence(series_data) -> Sequence:
    contour_image_sequence = Sequence()
    for series in series_data:
        contour_image = Dataset()
        contour_image.ReferencedSOPClassUID = series.SOPClassUID
        contour_image.ReferencedSOPInstanceUID = series.SOPInstanceUID
        contour_image_sequence.append(contour_image)
    return contour_image_sequence

def add_rs_series_information(ds: FileDataset, series_data):
    dt = datetime.datetime.now()
    ds.StructureSetDescription = dt.strftime("%Y%m%d") + "MBB"



"""Radiation Therapy Structure Set module"""

from .builder import create_rtstruct_dataset

from .add_new_roi import create_roi_into_rs_ds

from .make_contour_sequence import (
    add_contour_sequence_from_mask3d,
    add_contour_sequence_from_dcm_ctr_dict
)

from .parser import (
    get_contour_dict,
    get_roi_number_to_name
)

from .checker import (
    is_rtstruct_matching_series,
    check_rs_iod
)

from .rs_to_volume import (
    rtstruct_to_mask_dict,
    calc_image_series_affine_mapping,
    calc_rs_affine_mapping
)


__all__ = [
    'create_rtstruct_dataset',
    'create_roi_into_rs_ds',

    'add_contour_sequence_from_mask3d',
    'add_contour_sequence_from_dcm_ctr_dict',

    'get_contour_dict',
    'get_roi_number_to_name',

    'is_rtstruct_matching_series',
    'check_rs_iod',

    'rtstruct_to_mask_dict',
    'calc_image_series_affine_mapping',
    'calc_rs_affine_mapping',

]



from pydicom.uid import UID

# Deformable Spatial Registration CIOD
DEFORMABLE_SPATIAL_REGSITRATION_IOD = {
    "PatientID": {},
    "PatientName": {},
    "PatientBirthDate": {},
    "PatientSex": {},
    "StudyInstanceUID": {},
    "StudyID": {},
    "StudyDate": {},
    "StudyTime": {},
    "AccessionNumber": {},
    "SeriesInstanceUID": {"type": UID},
    "SeriesNumber": {},
    "Modality": {"value": "REG"},
    "Manufacturer": {},
    "FrameOfReferenceUID": {"type": UID},
    "SOPClassUID": {"type": UID},
    "SOPInstanceUID": {"type": UID},
    "InstanceNumber": {},
    "ContentLabel": {},
    "ContentDescription": {},
    "DeformableRegistrationSequence": {
        "SourceFrameOfReferenceUID": {"type": UID},
        "submap": {
            "DeformableRegistrationGridSequence": {
                "ImagePositionPatient": {},
                "ImageOrientationPatient": {},
                "GridDimensions": {},
                "GridResolution": {},
                "VectorGridData": {}
            },
            "PreDeformationMatrixRegistrationSequence": {},
            "PostDeformationMatrixRegistrationSequence": {},
            "RegistrationTypeCodeSequence": {},
        }
    },

}



"""
Check RTSTRUCT dataset Class Information Object Definition (IOD)

Author: Higumalu
Date: 2025-06-13
"""
from pydicom.dataset import Dataset

from pydicomrt.utils.validate_dcm_info import check_iod, ValidationError
from .rs_ds_iod import RT_STRUCTURE_SET_IOD


def contour_data_validator(value):
    if hasattr(value, '__len__'):
        if len(value) == 0:
            raise ValidationError("Contour data is empty")
        elif len(value) < 9:
            raise ValidationError("Contour data less then 3 points")
        elif len(value) % 3 != 0:
            raise ValidationError("Contour data length is not multiple of 3")
    else:
        raise ValidationError("Contour data must be a sequence type that supports len()")


RS_VALIDATORS_MAP = {
    "ContourDataValidator": contour_data_validator,
}


def check_rs_iod(rs_ds: Dataset) -> dict:
    result_dict = {
        "result": True,
        "content": []
    }
    failed_item_list = []
    failed_item_list = check_iod(ds=rs_ds, config_map=RT_STRUCTURE_SET_IOD, validators=RS_VALIDATORS_MAP, path="")

    if len(failed_item_list) > 0:
        result_dict["result"] = False
        result_dict["content"] = failed_item_list

    return result_dict

# --------------------------------------------------------------------------------------------------------------------- #

def is_rtstruct_matching_series(
    rs_ds: Dataset,
    series_ds_list: list) -> bool:
    """
    Check if the RTSTRUCT dataset is matching the series dataset
    :param rs_ds: RTSTRUCT dataset
    :param series_ds_list: list of series dataset
    :return: True if matching, False otherwise
    """
    status = 0
    first_slice = series_ds_list[0]
    if rs_ds.FrameOfReferenceUID == first_slice.FrameOfReferenceUID:
        status += 1

    try:
        rs_ds_series_uid_list = rs_ds.ReferencedFrameOfReferenceSequence[0].RTReferencedSeriesSequence[0].SeriesInstanceUID
        if rs_ds_series_uid_list == first_slice.SeriesInstanceUID:
            status += 1
    except Exception:
        pass

    try:
        sop_instance_uid_list = [ds.SOPInstanceUID for ds in series_ds_list]
        for roi_contour_sequence in rs_ds.ROIContourSequence:
            contour_sequence = roi_contour_sequence.ContourSequence
            for contour in contour_sequence:
                for contour_image_sequence in contour.ContourImageSequence:
                    if contour_image_sequence.ReferencedSOPInstanceUID not in sop_instance_uid_list:
                        status = 0
                        break
    except Exception:
        pass

    return status > 0



import numpy as np

from pydicom.dataset import Dataset


def get_slice_position(series_slice: Dataset):
    _, _, slice_direction = get_slice_directions(series_slice)
    return np.dot(slice_direction, series_slice.ImagePositionPatient)


def get_slice_directions(series_slice: Dataset):
    orientation = series_slice.ImageOrientationPatient
    row_direction = np.array(orientation[:3])
    column_direction = np.array(orientation[3:])
    slice_direction = np.cross(row_direction, column_direction)

    if not np.allclose(
        np.dot(row_direction, column_direction), 0.0, atol=1e-3
    ) or not np.allclose(np.linalg.norm(slice_direction), 1.0, atol=1e-3):
        raise Exception("Invalid Image Orientation (Patient) attribute")

    return row_direction, column_direction, slice_direction


def get_spacing_between_slices(series_data):
    if len(series_data) > 1:
        first = get_slice_position(series_data[0])
        last = get_slice_position(series_data[-1])
        return (last - first) / (len(series_data) - 1)

    # Return nonzero value for one slice just to make the transformation matrix invertible
    return 1.0


def get_pixel_to_patient_transformation_matrix(series_data):

    first_slice = series_data[0]

    offset = np.array(first_slice.ImagePositionPatient)
    row_spacing, column_spacing = first_slice.PixelSpacing
    slice_spacing = get_spacing_between_slices(series_data)
    row_direction, column_direction, slice_direction = get_slice_directions(first_slice)

    mat = np.identity(4, dtype=np.float32)
    mat[:3, 0] = row_direction * row_spacing
    mat[:3, 1] = column_direction * column_spacing
    mat[:3, 2] = slice_direction * slice_spacing
    mat[:3, 3] = offset     # + np.array([0.5, 0.5, 0.0])

    return mat


def get_patient_to_pixel_transformation_matrix(series_data):
    first_slice = series_data[0]

    offset = np.array(first_slice.ImagePositionPatient)
    row_spacing, column_spacing = first_slice.PixelSpacing
    slice_spacing = get_spacing_between_slices(series_data)
    row_direction, column_direction, slice_direction = get_slice_directions(first_slice)

    linear = np.identity(3, dtype=np.float32)
    linear[0, :3] = row_direction / row_spacing
    linear[1, :3] = column_direction / column_spacing
    linear[2, :3] = slice_direction / slice_spacing

    mat = np.identity(4, dtype=np.float32)
    mat[:3, :3] = linear
    mat[:3, 3] = offset.dot(-linear.T)

    return mat


def apply_transformation_to_3d_points(
    points: np.ndarray, transformation_matrix: np.ndarray
    ):
    """
    * Augment each point with a '1' as the fourth coordinate to allow translation
    * Multiply by a 4x4 transformation matrix
    * Throw away added '1's
    """
    vec = np.concatenate((points, np.ones((points.shape[0], 1))), axis=1)
    return vec.dot(transformation_matrix.T)[:, :3]



from pydicom.multival import MultiValue
from pydicom.uid import UID

RT_STRUCTURE_SET_IOD = {
    "PatientID": {},
    "PatientName": {},
    "PatientBirthDate": {},
    "PatientSex": {},
    "StudyInstanceUID": {},
    "StudyID": {},
    "StudyDate": {},
    "StudyTime": {},
    "AccessionNumber": {},
    "SeriesInstanceUID": {"type": UID},
    "SeriesNumber": {},
    "OperatorsName": {},
    "Modality": {},
    "Manufacturer": {},
    "FrameOfReferenceUID": {"type": UID},
    "SOPClassUID": {"type": UID},
    "SOPInstanceUID": {"type": UID},
    "StructureSetLabel": {"type": str},
    "StructureSetDate": {},
    "StructureSetTime": {},
    "StructureSetROISequence": {
        "submap": {
            "ROINumber": {"type": int},
            "ROIName": {"type": str},
            "ROIGenerationAlgorithm": {},
            "ReferencedFrameOfReferenceUID": {},
        }
    },
    "ROIContourSequence": {
        "submap": {
            "ReferencedROINumber": {"type": int},
            "ROIDisplayColor": {},
            "ContourSequence": {
                "optional": True,
                "submap": {
                    "ContourImageSequence": {
                        "submap": {
                            "ReferencedSOPClassUID": {"type": UID},
                            "ReferencedSOPInstanceUID": {"type": UID},
                        }
                    },
                    "ContourGeometricType": {},
                    "NumberOfContourPoints": {"type": int},
                    "ContourData": {
                        "type": MultiValue,
                        "validator": ["ContourDataValidator"]
                        }
                }
            }
        }
    },
    "RTROIObservationsSequence": {
        "submap": {
            "ReferencedROINumber": {"type": int},
            "ObservationNumber": {"type": int},
            "RTROIInterpretedType": {},
        }
    }
}



from pydicom.uid import UID


CT_IMAGE_IOD = {
    "PatientID": {},
    "PatientName": {},
    "PatientBirthDate": {},
    "PatientSex": {},
    "StudyInstanceUID": {},
    "StudyID": {},
    "StudyDate": {},
    "StudyTime": {},
    "AccessionNumber": {},
    "SeriesInstanceUID": {"type": UID},
    "SeriesNumber": {},
    "Modality": {"value": "CT"},
    "Manufacturer": {},
    "FrameOfReferenceUID": {"type": UID},
    "SOPClassUID": {"type": UID},
    "SOPInstanceUID": {"type": UID},
    "InstanceNumber": {},

    "SliceThickness": {},
    "ImagePositionPatient": {},
    "ImageOrientationPatient": {},
    "PixelSpacing": {},

    "Rows": {},
    "Columns": {},

    "ImageType": {},
    "BitsAllocated": {},
    "BitsStored": {},
    "HighBit": {},
    "RescaleIntercept": {},
    "RescaleSlope": {},

    "PixelData": {},

}



import SimpleITK as sitk


def rigid_registration(
    fixed_image: sitk.Image,
    moving_image: sitk.Image,
    histogram_bins: int = 100,
    learning_rate: float = 2.0,
    iterations: int = 100,
    convergence_minimum_value: float = 1e-6,
    convergence_window_size: int = 10,
    ) -> sitk.Transform:
    """
    Perform rigid registration between two 3D images using SimpleITK.

    This function aligns a moving image to a fixed image by estimating
    a rigid transformation (translation + rotation, no scaling/shearing).
    The method uses Mattes Mutual Information as a similarity metric,
    a gradient descent optimizer, and returns an affine transform
    containing only the rigid components.

    Args:
        fixed_image (sitk.Image): The reference image to which the moving image is aligned.
        moving_image (sitk.Image): The image to be registered (transformed).
        histogram_bins (int, optional): Number of histogram bins for Mattes Mutual Information metric. Default = 50.
        learning_rate (float, optional): Step size for gradient descent optimizer. Default = 1.0.
        iterations (int, optional): Maximum number of optimizer iterations. Default = 100.
        convergence_minimum_value (float, optional): Minimum convergence value for optimizer stopping criterion. Default = 1e-6.
        convergence_window_size (int, optional): Window size for convergence checking. Default = 10.

    Returns:
        sitk.Transform: An affine transform containing only rotation and translation
                        that aligns the moving image to the fixed image.
    """

    # Ensure both images are float32 for numerical stability in registration
    fixed_image = sitk.Cast(fixed_image, sitk.sitkFloat32)
    moving_image = sitk.Cast(moving_image, sitk.sitkFloat32)

    # Create the registration method object
    registration_method = sitk.ImageRegistrationMethod()

    # Use Mattes Mutual Information (robust for multimodal image registration)
    registration_method.SetMetricAsMattesMutualInformation(numberOfHistogramBins=histogram_bins)

    # Use linear interpolation for resampling the moving image
    registration_method.SetInterpolator(sitk.sitkLinear)

    # Configure the optimizer as gradient descent with given parameters
    registration_method.SetOptimizerAsGradientDescent(
        learningRate=learning_rate,
        numberOfIterations=iterations,
        convergenceMinimumValue=convergence_minimum_value,
        convergenceWindowSize=convergence_window_size,
    )

    # Scale optimizer step sizes according to physical units of the image
    registration_method.SetOptimizerScalesFromPhysicalShift()

    # Initialize with a centered rigid transform (rotation + translation)
    initial_transform = sitk.CenteredTransformInitializer(
        fixed_image,
        moving_image,
        sitk.VersorRigid3DTransform(),
        sitk.CenteredTransformInitializerFilter.GEOMETRY
    )

    # Explicitly set the center to (0,0,0) – prevents unwanted shifting
    initial_transform.SetCenter((0.0, 0.0, 0.0))

    # Assign the initial transform to the registration method
    registration_method.SetInitialTransform(initial_transform, inPlace=False)

    # Run the registration (this optimizes the transform parameters)
    final_transform = registration_method.Execute(fixed_image, moving_image)

    # If the result is a composite transform, extract the first (rigid) transform
    if isinstance(final_transform, sitk.CompositeTransform):
        transform = final_transform.GetNthTransform(0)
    else:
        transform = final_transform

    # Convert rigid transform into an affine transform (matrix + translation only)
    affine_transform = sitk.AffineTransform(3)
    affine_transform.SetMatrix(transform.GetMatrix())       # Rotation part
    affine_transform.SetTranslation(transform.GetTranslation())  # Translation part

    return affine_transform


def alignment_registration(fixed_image, moving_image, moments=True, interpolator=sitk.sitkLinear, default_value=0):
    """
    A simple registration procedure that can align images in a single step.
    Uses the image centres-of-mass (and optionally second moments) to
    estimate the shift (and rotation) needed for alignment.

    Args:
        fixed_image ([SimpleITK.Image]): The fixed (target/primary) image.
        moving_image ([SimpleITK.Image]): The moving (secondary) image.
        moments (bool, optional): Option to align images using the second moment. Defaults to True.
        interpolator (int, optional): The interpolation order.
                                Available options:
                                    - SimpleITK.sitkNearestNeighbor
                                    - SimpleITK.sitkLinear
                                    - SimpleITK.sitkBSpline
                                Defaults to SimpleITK.sitkLinear.
        default_value (int, optional): Default (background) value. Defaults to 0.

    Returns:
        [SimpleITK.Image]: The registered moving (secondary) image.
        [SimleITK.Transform]: The linear transformation.
    """

    moving_image_type = moving_image.GetPixelIDValue()
    fixed_image = sitk.Cast(fixed_image, sitk.sitkFloat32)
    moving_image = sitk.Cast(moving_image, sitk.sitkFloat32)
    initial_transform = sitk.CenteredTransformInitializer(
        fixed_image, moving_image, sitk.VersorRigid3DTransform(), moments
    )
    aligned_image = sitk.Resample(moving_image, fixed_image, initial_transform, interpolator, default_value)
    aligned_image = sitk.Cast(aligned_image, moving_image_type)
    return aligned_image, initial_transform


def registration_command_iteration(method):
    """
    Utility function to print information during (rigid, similarity, translation, B-splines)
    registration
    """
    print("{0:3} = {1:10.5f}".format(method.GetOptimizerIteration(), method.GetMetricValue()))


def apply_transform(
    input_image,
    reference_image=None,
    transform=None,
    default_value=0,
    interpolator=sitk.sitkNearestNeighbor,
):
    """
    Transform a volume of structure with the given deformation field.

    Args
        input_image (SimpleITK.Image): The image, to which the transform is applied
        reference_image (SimpleITK.Image): The image will be resampled into this reference space.
        transform (SimpleITK.Transform): The transformation
        default_value: Default (background) value. Defaults to 0.
        interpolator (int, optional): The interpolation order.
                                Available options:
                                    - SimpleITK.sitkNearestNeighbor
                                    - SimpleITK.sitkLinear
                                    - SimpleITK.sitkBSpline
                                Defaults to SimpleITK.sitkNearestNeighbor

    Returns
        (SimpleITK.Image): the transformed image

    """
    original_image_type = input_image.GetPixelID()

    resampler = sitk.ResampleImageFilter()

    if reference_image:
        resampler.SetReferenceImage(reference_image)
    else:
        resampler.SetReferenceImage(input_image)

    if transform:
        resampler.SetTransform(transform)

    resampler.SetDefaultPixelValue(default_value)
    resampler.SetInterpolator(interpolator)

    output_image = resampler.Execute(input_image)
    output_image = sitk.Cast(output_image, original_image_type)

    return output_image


def linear_registration(
    fixed_image,
    moving_image,
    fixed_structure=None,
    moving_structure=None,
    reg_method="similarity",
    metric="mean_squares",
    optimiser="gradient_descent",
    shrink_factors=[8, 2, 1],
    smooth_sigmas=[4, 2, 0],
    sampling_rate=0.25,
    final_interp=2,
    number_of_iterations=100,
    default_value=None,
    verbose=False,
):
    """
    Initial linear registration between two images.
    The images are not required to be in the same space.
    There are several transforms available, with options for the metric and optimiser to be used.
    Note the default_value, which should be set to match the image modality.

    Args:
        fixed_image ([SimpleITK.Image]): The fixed (target/primary) image.
        moving_image ([SimpleITK.Image]): The moving (secondary) image.
        fixed_structure (bool, optional): If defined, a binary SimpleITK.Image used to mask metric
                                          evaluation for the moving image. Defaults to False.
        moving_structure (bool, optional): If defined, a binary SimpleITK.Image used to mask metric
                                           evaluation for the fixed image. Defaults to False.
        reg_method (str, optional): The linear transformtation model to be used for image
                                    registration.
                                    Available options:
                                     - translation
                                     - rigid
                                     - similarity
                                     - affine
                                     - scale
                                     - scaleversor
                                     - scaleskewversor
                                    Defaults to "Similarity".
        metric (str, optional): The metric to be optimised during image registration.
                                Available options:
                                 - correlation
                                 - mean_squares
                                 - mattes_mi
                                 - joint_hist_mi
                                Defaults to "mean_squares".
        optimiser (str, optional): The optimiser algorithm used for image registration.
                                   Available options:
                                    - lbfgsb
                                      (limited-memory Broyden–Fletcher–Goldfarb–Shanno (bounded).)
                                    - gradient_descent
                                    - gradient_descent_line_search
                                   Defaults to "gradient_descent".
        shrink_factors (list, optional): The multi-resolution downsampling factors.
                                         Defaults to [8, 2, 1].
        smooth_sigmas (list, optional): The multi-resolution smoothing kernel scale (Gaussian).
                                        Defaults to [4, 2, 0].
        sampling_rate (float, optional): The fraction of voxels sampled during each iteration.
                                         Defaults to 0.25.
        ants_radius (int, optional): Used is the metric is set as "ants_radius". Defaults to 3.
        final_interp (int, optional): The final interpolation order. Defaults to 2 (linear).
        number_of_iterations (int, optional): Number of iterations in each multi-resolution step.
                                              Defaults to 50.
        default_value (int, optional): Default voxel value. Defaults to 0 unless image is CT-like.
        verbose (bool, optional): Print image registration process information. Defaults to False.

    Returns:
        [SimpleITK.Image]: The registered moving (secondary) image.
        [SimleITK.Transform]: The linear transformation.
    """

    # Re-cast
    fixed_image = sitk.Cast(fixed_image, sitk.sitkFloat32)

    moving_image_type = moving_image.GetPixelIDValue()
    moving_image = sitk.Cast(moving_image, sitk.sitkFloat32)

    # Initialise using a VersorRigid3DTransform
    initial_transform = sitk.CenteredTransformInitializer(
        fixed_image, moving_image, sitk.Euler3DTransform(), False
    )
    # Set up image registration method
    registration = sitk.ImageRegistrationMethod()

    registration.SetShrinkFactorsPerLevel(shrink_factors)
    registration.SetSmoothingSigmasPerLevel(smooth_sigmas)
    registration.SmoothingSigmasAreSpecifiedInPhysicalUnitsOn()

    registration.SetMovingInitialTransform(initial_transform)

    if metric.lower() == "correlation":
        registration.SetMetricAsCorrelation()
    elif metric.lower() == "mean_squares":
        registration.SetMetricAsMeanSquares()
    elif metric.lower() == "mattes_mi":
        registration.SetMetricAsMattesMutualInformation()
    elif metric.lower() == "joint_hist_mi":
        registration.SetMetricAsJointHistogramMutualInformation()
    # to do: add the rest

    registration.SetInterpolator(sitk.sitkLinear)  # Perhaps a small gain in improvement
    registration.SetMetricSamplingPercentage(sampling_rate, seed=42)
    registration.SetMetricSamplingStrategy(sitk.ImageRegistrationMethod.REGULAR)

    # This is only necessary if using a transform comprising changes with different units
    # e.g. rigid (rotation: radians, translation: mm)
    # It can safely be left on
    registration.SetOptimizerScalesFromPhysicalShift()

    if moving_structure:
        registration.SetMetricMovingMask(moving_structure)

    if fixed_structure:
        registration.SetMetricFixedMask(fixed_structure)

    if isinstance(reg_method, str):
        if reg_method.lower() == "translation":
            registration.SetInitialTransform(sitk.TranslationTransform(3))
        elif reg_method.lower() == "similarity":
            registration.SetInitialTransform(sitk.Similarity3DTransform())
        elif reg_method.lower() == "affine":
            registration.SetInitialTransform(sitk.AffineTransform(3))
        elif reg_method.lower() == "rigid":
            registration.SetInitialTransform(sitk.VersorRigid3DTransform())
        elif reg_method.lower() == "scale":
            registration.SetInitialTransform(sitk.ScaleTransform(3))
        elif reg_method.lower() == "scaleversor":
            registration.SetInitialTransform(sitk.ScaleVersor3DTransform())
        elif reg_method.lower() == "scaleskewversor":
            registration.SetInitialTransform(sitk.ScaleSkewVersor3DTransform())
        else:
            raise ValueError(
                "You have selected a registration method that does not exist.\n Please select from"
                " Translation, Similarity, Affine, Rigid, ScaleVersor, ScaleSkewVersor"
            )
    elif isinstance(
        reg_method,
        (
            sitk.CompositeTransform,
            sitk.Transform,
            sitk.TranslationTransform,
            sitk.Similarity3DTransform,
            sitk.AffineTransform,
            sitk.VersorRigid3DTransform,
            sitk.ScaleVersor3DTransform,
            sitk.ScaleSkewVersor3DTransform,
        ),
    ):
        registration.SetInitialTransform(reg_method)
    else:
        raise ValueError(
            "'reg_method' must be either a string (see docs for acceptable registration names), "
            "or a custom sitk.CompositeTransform."
        )

    if optimiser.lower() == "lbfgsb":
        registration.SetOptimizerAsLBFGSB(
            gradientConvergenceTolerance=1e-5,
            numberOfIterations=number_of_iterations,
            maximumNumberOfCorrections=50,
            maximumNumberOfFunctionEvaluations=1024,
            costFunctionConvergenceFactor=1e7,
            trace=verbose,
        )
    elif optimiser.lower() == "exhaustive":
        """
        This isn't well implemented
        Needs some work to give options for sampling rates
        Use is not currently recommended
        """
        samples = [10, 10, 10, 10, 10, 10]
        registration.SetOptimizerAsExhaustive(samples)
    elif optimiser.lower() == "gradient_descent_line_search":
        registration.SetOptimizerAsGradientDescentLineSearch(
            learningRate=1.0, numberOfIterations=number_of_iterations
        )
    elif optimiser.lower() == "gradient_descent":
        registration.SetOptimizerAsGradientDescent(
            learningRate=2.0, numberOfIterations=number_of_iterations
        )

    if verbose:
        registration.AddCommand(
            sitk.sitkIterationEvent,
            lambda: registration_command_iteration(registration),
        )

    output_transform = registration.Execute(fixed=fixed_image, moving=moving_image)
    # Combine initial and optimised transform
    combined_transform = sitk.CompositeTransform([initial_transform, output_transform])

    # Try to find default value
    if default_value is None:
        default_value = 0

        # Test if image is CT-like
        if sitk.GetArrayViewFromImage(moving_image).min() <= -1000:
            default_value = -1000

    registered_image = apply_transform(
        input_image=moving_image,
        reference_image=fixed_image,
        transform=combined_transform,
        default_value=default_value,
        interpolator=final_interp,
    )

    registered_image = sitk.Cast(registered_image, moving_image_type)

    return registered_image, combined_transform


from pydicom.uid import UID

# Spatial Registration CIOD
SPATIAL_REGSITRATION_IOD = {
    "PatientID": {},
    "PatientName": {},
    "PatientBirthDate": {},
    "PatientSex": {},
    "StudyInstanceUID": {},
    "StudyID": {},
    "StudyDate": {},
    "StudyTime": {},
    "AccessionNumber": {},
    "SeriesInstanceUID": {"type": UID},
    "SeriesNumber": {},
    "Modality": {"value": "REG"},
    "Manufacturer": {},
    "FrameOfReferenceUID": {"type": UID},
    "SOPClassUID": {"type": UID},
    "SOPInstanceUID": {"type": UID},
    "InstanceNumber": {},
    "ContentLabel": {},
    "ContentDescription": {},
    "RegistrationSequence": {
        "FrameOfReferenceUID": {"type": UID},
        "submap": {
            "MatrixRegistrationSequence": {
                "submap": {
                    "MatrixSequence": {
                        "FrameOfReferenceTransformationMatrixType": {},
                        "FrameOfReferenceTransformationMatrix": {},
                    }
                }
            },
        }
    },

}



import numpy as np
import cv2

from pydicom.dataset import Dataset
from pydicom.sequence import Sequence

from utils.coordinate_transform import apply_transformation_to_3d_points
from .contour_process_method import contour_process


def add_mask3d_into_rsds(
    rs_ds,
    mask_volume,
    affine_map,
    roi_color,
    roi_number,
    roi_name,
    roi_description,
    roi_interpreted_type="ORGAN",
    contour_config: dict = {
        "ex_noise_size": 0,
        "in_noise_size": 0,
        "lowpass_ratio": 10,
        "ctr_precision": 8
        }
    ) -> Dataset:
    rs_ds.StructureSetROISequence.append(create_structure_set_roi(roi_number, roi_name, roi_description))
    rs_ds.ROIContourSequence.append(create_roi_contour_sequence(roi_number, roi_color, mask_volume, affine_map, contour_config))
    rs_ds.RTROIObservationsSequence.append(create_rtroi_observation(roi_number, roi_interpreted_type))
    return rs_ds

# StructureSetROISequence  --> [ROINumber, ReferencedFrameOfReferenceUID, ROIName, ROIGenerationAlgorithm=AUTOMATIC]
################################ StructureSetROISequence ####################################################################
def create_structure_set_roi(roi_number,
                             roi_name, roi_description,
                             roi_generation_algorithm="AUTOMATIC",
                             roi_generation_description="KumaGenerated") -> Dataset:
    # Structure Set ROI Sequence: Structure Set ROI 1
    structure_set_roi = Dataset()
    structure_set_roi.ROINumber = roi_number
    structure_set_roi.ROIName = roi_name
    structure_set_roi.ROIDescription = roi_description
    structure_set_roi.ROIGenerationAlgorithm = roi_generation_algorithm
    structure_set_roi.ROIGenerationDescription = roi_generation_description
    return structure_set_roi


# create ROIContourSequence -> ContourSequence -> [ContourImageSequence / ContourData / ContourNumber / ...]
################################ ROIContourSequence ####################################################################
def create_roi_contour_sequence(roi_number, roi_color, mask_volume, affine_map, ctr_config) -> Dataset:     # == create_roi_contour() in rt-utils
    roi_contour_sequence = Dataset()
    roi_contour_sequence.ReferencedROINumber = roi_number    # also in create_structure_set_roi
    roi_contour_sequence.ROIDisplayColor = roi_color
    roi_contour_sequence.ContourSequence = create_contour_sequence(mask_volume, affine_map, ctr_config)
    return roi_contour_sequence


def create_roi_contour_sequence_from_ctr_dict(roi_number, roi_color, ctr_dict, affine_map) -> Dataset:     # == create_roi_contour() in rt-utils
    roi_contour_sequence = Dataset()
    roi_contour_sequence.ReferencedROINumber = roi_number    # also in create_structure_set_roi
    roi_contour_sequence.ROIDisplayColor = roi_color
    roi_contour_sequence.ContourSequence = create_contour_sequence_from_ctr_dict(ctr_dict, affine_map)
    return roi_contour_sequence


# 3Dmask -[for loop]-> 2Dmask -[cv2.findContours]-> poly_list -[for loop]-> 2Dcontour
def create_contour_sequence(mask_volume, affine_map, ctr_config) -> Sequence:
    contour_sequence = Sequence()
    contour_data_seq = volume_to_contour_list(mask_volume, affine_map, ctr_config)
    for index, contour_data in enumerate(contour_data_seq):
        contour = create_contour_sequence_block(contour_data, precision=ctr_config['ctr_precision'])
        contour_sequence.append(contour)
    return contour_sequence


def create_contour_sequence_from_ctr_dict(ctr_dict, affine_map) -> Sequence:
    contour_sequence = Sequence()
    for z_key, polys in ctr_dict.items():
        for poly in polys:
            contour = Dataset()
            z_col = np.full((poly.shape[0], 1), z_key)
            contour_data = np.hstack((poly, z_col))
            contour_data = apply_transformation_to_3d_points(contour_data, affine_map)
            contour_data = list(contour_data.reshape(-1))
            contour.ContourData = contour_data
            contour.NumberOfContourPoints = (len(contour_data) / 3)
            contour.ContourGeometricType = "CLOSED_PLANAR"
            contour_sequence.append(contour)
    return contour_sequence


# ContourSequence -> [ContourImageSequence, ContourGeometricType, NumberOfContourPoints, ContourNumber, ContourData]
# -> ContourImageSequence -> [ReferencedSOPClassUID, ReferencedSOPInstanceUID]
def create_contour_sequence_block(contour_data, precision=4) -> Dataset:
    # contour_image = Dataset()
    # contour_image.ReferencedSOPClassUID = image_slice.SOPClassUID
    # contour_image.ReferencedSOPInstanceUID = image_slice.SOPInstanceUID
    # contour_image_sequence = Sequence()
    # contour_image_sequence.append(contour_image)

    contour = Dataset()
    # contour.ContourImageSequence = contour_image_sequence
    contour.ContourGeometricType = "CLOSED_PLANAR"
    contour.NumberOfContourPoints = (len(contour_data) / 3)
    contour_data = [round(coord, precision) for coord in contour_data]
    contour.ContourData = contour_data

    return contour

# 3Dmask
# -[for loop with ds_list]-> 2Dmask with series slice
# -[cv2.findContours]-> cv2_contours
# -[for loop]-> contour(polygon)
# -[smooth_ctr (maybe)]-> contour
# -[coordination transform]-> dicom fromat contour (ContourData)
def volume_to_contour_list(mask_volume, affine_map, ctr_config):
    formatted_contours = []
    for index, mask_slice in enumerate(mask_volume):
        mask_slice = mask_volume[index, :, :]           # mask_volume.shape = (z, y, x)     ex: (144, 512, 512)

        if np.sum(mask_slice) == 0:
            continue

        cv2_contours, hierarchy = cv2.findContours(
            mask_slice.astype(np.uint8),
            cv2.RETR_CCOMP,
            cv2.CHAIN_APPROX_NONE
            )
        if cv2_contours is None or hierarchy is None: continue

        for hier_index, contour in enumerate(cv2_contours):
            poly_hierarchy = hierarchy[0][hier_index][3]

            x_points, y_points = contour.T[:, 0]
            # interface for contour smoothing and other process
            x_points, y_points = contour_process(
                x_points, y_points, poly_hierarchy,
                external_noise_size=ctr_config["ex_noise_size"],
                internal_noise_size=ctr_config["in_noise_size"],
                low_pass_ratio=ctr_config["lowpass_ratio"]
                )
            # check if contour is not empty
            if len(x_points) < 3 or len(y_points) < 3: continue

            contour = np.stack([x_points, y_points]).T
            contour = np.concatenate((np.array(contour), np.full((len(contour), 1), index)), axis=1)
            # transformation_matrix = get_pixel_to_patient_transformation_matrix(image_series)
            transformed_contour = apply_transformation_to_3d_points(contour, affine_map)

            dicom_formatted_contour = np.ravel(transformed_contour).tolist()
            formatted_contours.append(dicom_formatted_contour)

    return formatted_contours

# ------------------------------- RTROIObservationsSequence --------------------------------------- #
def create_rtroi_observation(roi_number, roi_interpreted_type='ORGAN') -> Dataset:
    rtroi_observation = Dataset()
    rtroi_observation.ObservationNumber = roi_number
    rtroi_observation.ReferencedROINumber = roi_number

    rtroi_observation.ROIObservationDescription = "Type:Soft,Range:*/*,Fill:0,Opacity:0.0,Thickness:1,LineThickness:2,read-only:false"
    rtroi_observation.private_creators = "higumalu"
    rtroi_observation.RTROIInterpretedType = roi_interpreted_type
    rtroi_observation.ROIInterpreter = ""
    return rtroi_observation



"""
Check Dicom Class Information Object Definition (IOD)

Author: Higumalu
Date: 2025-06-13
"""
from typing import Callable, Dict, List
from pydicom.tag import Tag
from pydicom.sequence import Sequence
from pydicom.datadict import tag_for_keyword

class ValidationError(Exception):
    pass

def resolve_tag(key):
    if isinstance(key, str) and not key.lower().startswith("0x"):
        tag = tag_for_keyword(key)
        if tag is None:
            raise ValueError(f"Unknown DICOM keyword: {key}")
        return tag
    return Tag(key)

def check_iod(
    ds,
    config_map: Dict[str, dict],
    validators: Dict[str, Callable] = None,
    path: str = ""
    ) -> List[str]:
    """
    Check if a DICOM dataset conforms to a given IOD configuration.
    :param ds: The DICOM dataset to check.
    :param config_map: The IOD configuration to check against.
        {
        "tag_name": {
            "type": type,
            "validator": ["validator_name_1", "validator_name_2"],
            "submap": {
                "tag_name": {
                    "type": type,
                    "validator": ["validator_name_1", "validator_name_2"],
                    "submap": {
                        ...
                    }
                }
            }
        }
    :param validators: A dictionary of custom validators.
        {
        "validator_name_1": validator_function_1,
        "validator_name_2": validator_function_2,
        ...
        }
    :param path: The path to the dataset in the IOD configuration.
    :return: A list of errors.
    """
    # print(validators)
    errors = []
    for key, cfg in config_map.items():
        tag = resolve_tag(key)
        elem = ds.get(tag, None)
        loc = path or "root"

        if elem is None:
            if cfg.get("optional", False):
                continue
            else:
                errors.append(f"Missing {key} in {loc}")
                continue

        val = elem.value

        # Check type
        expected_type = cfg.get("type")
        if expected_type and not isinstance(val, expected_type):
            errors.append(f"{key} in {loc} should be {expected_type.__name__}, got {type(val).__name__} {val}")

        # Check value
        expected_value = cfg.get("value")
        if expected_value and val != expected_value:
            errors.append(f"{key} in {loc} should be {expected_value}, got {val}")

        # ------------------------------ Custom validator --------------------------------------- #
        for vname in cfg.get("validator", []):
            func = validators.get(vname) if validators else None
            if func is None:
                # raise ValueError(f"Unknown validator '{vname}', provide a dict of validators")
                errors.append(f"No validator named '{vname}' provided for {key} in {loc}, provide a dict of validators")
                continue
            try:
                func(val)
            except ValidationError as e:
                errors.append(f"{key} in {loc} failed '{vname}': {e}")
        # ---------------------------------------------------------------------------------------- #

        # If there is a submap, recursively check the Sequence
        submap = cfg.get("submap")
        if submap:
            if elem.VR != "SQ" or not isinstance(val, Sequence):
                errors.append(f"{key} in {loc} should be SQ, but VR={elem.VR}")
            else:
                for idx, item in enumerate(val):
                    errors += check_iod(item, submap, path=f"{loc}.{key}[{idx}]", validators=validators)

    return errors


if __name__ == "__main__":
    # Example validator function
    def example_string_len_16_validator(value):
        if len(value) > 16:
            raise ValidationError(f"String length should be <= 16, but got {len(value)}")

    EXAMPLE_VAILDATORS = {
        "string_len_16": example_string_len_16_validator,
    }

    EXAMPLE_CONFIG = {
        "PatientName": {
            "type": str,
            "validator": ["string_len_16"],
        },
    }




import numpy as np
import SimpleITK as sitk

def get_dose_array_spacing(dose_ds):
    x_spacing, y_spacing = dose_ds.PixelSpacing
    frame_offset_vector_list = dose_ds.GridFrameOffsetVector
    frame_distance_list = [np.linalg.norm(frame_offset_vector_list[i + 1] - frame_offset_vector_list[i])
                           for i in range(len(frame_offset_vector_list) - 1)]
    z_spacing = np.mean(frame_distance_list)
    std_distance = np.std(frame_distance_list)
    if std_distance > 0.001:
        print("Warning: dose array spacing is not uniform")
    return [x_spacing, y_spacing, z_spacing]

def get_dose_direction(dose_ds):
    ori = dose_ds.ImageOrientationPatient
    row = ori[:3]
    col = ori[3:]
    cross = np.cross(row, col)
    dose_direction = np.asarray([row, col, cross]).T
    return dose_direction

def get_dose_origin(dose_ds):
    return dose_ds.ImagePositionPatient

def get_dose_array(dose_ds):
    return dose_ds.pixel_array * dose_ds.DoseGridScaling

def get_dose_sitk_image(dose_ds):
    dose_array = get_dose_array(dose_ds)
    spacing = get_dose_array_spacing(dose_ds)
    origin = get_dose_origin(dose_ds)
    direction = get_dose_direction(dose_ds)
    sitk_image = sitk.GetImageFromArray(dose_array)
    sitk_image.SetSpacing(spacing)
    sitk_image.SetOrigin(origin)
    sitk_image.SetDirection(direction.flatten())
    return sitk_image



from .image_series_loader import load_sorted_image_series
from .coordinate_transform import apply_transformation_to_3d_points, get_pixel_to_patient_transformation_matrix
from .validate_dcm_info import check_iod

__all__ = [
    "load_sorted_image_series",

    "get_pixel_to_patient_transformation_matrix",
    "apply_transformation_to_3d_points",

    "check_iod",
]



import SimpleITK as sitk
import numpy as np
from typing import Optional


def _infer_background_value(img: sitk.Image, default_fallback: float = 0.0) -> float:
    """If the image looks like a CT (min <= -1000), return -1000. Otherwise, return fallback."""
    try:
        arr_min = float(np.min(sitk.GetArrayViewFromImage(img)))
        return -1000.0 if arr_min <= -1000.0 else float(default_fallback)
    except Exception:
        return float(default_fallback)

def _ensure_binary_mask(mask: sitk.Image) -> sitk.Image:
    """Ensure binary mask (0/1)."""
    return sitk.Cast(mask > 0, sitk.sitkUInt8)

def _soft_weight_from_mask(mask: sitk.Image, falloff_mm: float = 8.0) -> sitk.Image:
    """
    Generate soft weights (0..1):
    - Inside ROI, far from boundary → ~1
    - Near boundary → 0
    - Outside ROI → 0

    Uses SignedMaurerDistanceMap with linear falloff in mm.
    """
    mask = _ensure_binary_mask(mask)
    dt = sitk.SignedMaurerDistanceMap(
        mask, squaredDistance=False, insideIsPositive=True, useImageSpacing=True
    )
    w = sitk.Clamp(dt / float(falloff_mm), lowerBound=0.0, upperBound=1.0)
    return sitk.Cast(w, sitk.sitkFloat32)

def _apply_soft_mask(img: sitk.Image, w: sitk.Image, background_value: float) -> sitk.Image:
    """Apply soft mask: weighted_img = w * img + (1-w) * C."""
    img_f = sitk.Cast(img, sitk.sitkFloat32)
    return w * img_f + (1.0 - w) * float(background_value)

def _resample_to_reference(
    img: sitk.Image,
    reference: sitk.Image,
    transform: sitk.Transform = sitk.Transform(),
    interp: int = sitk.sitkLinear,
    default_value: float = 0.0,
    pixel_id: Optional[int] = None,
) -> sitk.Image:
    """Resample `img` onto the reference grid (direction, spacing, size, origin)."""
    if pixel_id is None:
        pixel_id = img.GetPixelID()
    return sitk.Resample(img, reference, transform, interp, default_value, pixel_id)


def demons_with_soft_mask(
    fixed: sitk.Image,
    moving: sitk.Image,
    demons_fn,  # demons_registration(fixed, moving, **kwargs)
    *,
    fixed_mask: sitk.Image,
    moving_mask: Optional[sitk.Image] = None,
    fixed_falloff_mm: float = 8.0,
    moving_falloff_mm: Optional[float] = None,
    background_fixed: Optional[float] = None,
    background_moving: Optional[float] = None,
    propagate_fixed_weight_to_moving: bool = False,
    initial_transform: Optional[sitk.Transform] = None,
    interp_order: int = sitk.sitkLinear,
    **demons_kwargs,
):
    """
    Run Demons registration with soft mask weighting, focusing on ROI.

    - `fixed_mask` is required; `moving_mask` is optional.
    - If `propagate_fixed_weight_to_moving=True`, the fixed weight map (w_f) will be
      warped to moving space using `initial_transform` and used as the moving weight.
    - If both `moving_mask` and propagation are provided, their minimum is used as the moving weight.
    """

    # 1) Weight field (fixed image)
    if fixed_mask is None:
        raise ValueError("fixed_mask is required.")
    w_f = _soft_weight_from_mask(fixed_mask, falloff_mm=float(fixed_falloff_mm))  # in fixed space

    # 2) Weight field (moving image)
    w_m_list = []

    # 2a) From moving_mask (in moving space)
    if moving_mask is not None:
        mf = float(moving_falloff_mm) if moving_falloff_mm is not None else float(fixed_falloff_mm)
        w_m_from_own = _soft_weight_from_mask(moving_mask, falloff_mm=mf)  # in moving space
        w_m_list.append(w_m_from_own)

    # 2b) Propagate fixed weight to moving space (if requested)
    if propagate_fixed_weight_to_moving:
        # Need to map fixed weights (w_f) into moving space:
        # moving(x) corresponds to fixed(T(x)) → weight = w_f(T(x))
        # If initial_transform maps moving→fixed, we can directly resample w_f.
        T = initial_transform if initial_transform is not None else sitk.Transform()
        w_f_on_m = _resample_to_reference(
            w_f, reference=moving, transform=T, interp=sitk.sitkLinear, default_value=0.0, pixel_id=sitk.sitkFloat32
        )
        w_m_list.append(w_f_on_m)

    # 2c) Combine moving weights: use min if multiple sources exist
    if len(w_m_list) == 1:
        w_m = w_m_list[0]
    elif len(w_m_list) > 1:
        w_m = sitk.Minimum(w_m_list[0], w_m_list[1]) if len(w_m_list) == 2 else w_m_list[0]
        for i in range(2, len(w_m_list)):
            w_m = sitk.Minimum(w_m, w_m_list[i])
    else:
        w_m = None  # No moving weight → no soft mask applied to moving

    # 3) Infer background values
    if background_fixed is None:
        background_fixed = _infer_background_value(fixed, default_fallback=0.0)
    if background_moving is None:
        background_moving = _infer_background_value(moving, default_fallback=0.0)

    # 4) Apply soft mask to images
    fixed_soft = _apply_soft_mask(fixed, w_f, background_fixed)
    moving_soft = _apply_soft_mask(moving, w_m, background_moving) if w_m is not None else moving

    # 5) Call user-provided Demons function
    #    This may internally perform multi-resolution pyramids and resampling.
    #    Pass initial_transform, interp_order, and other parameters downstream if supported.
    result_img, out_tfm, dvf = demons_fn(
        fixed_image=fixed_soft,
        moving_image=moving_soft,
        initial_displacement_field=demons_kwargs.pop("initial_displacement_field", None),
        isotropic_resample=demons_kwargs.pop("isotropic_resample", False),
        resolution_staging=demons_kwargs.pop("resolution_staging", (8, 4, 1)),
        iteration_staging=demons_kwargs.pop("iteration_staging", (20, 20, 20)),
        regularisation_kernel_mm=demons_kwargs.pop("regularisation_kernel_mm", 1.0),
        smoothing_sigma_factor=demons_kwargs.pop("smoothing_sigma_factor", 1.0),
        smoothing_sigmas=demons_kwargs.pop("smoothing_sigmas", False),
        default_value=demons_kwargs.pop("default_value", background_moving),
        ncores=demons_kwargs.pop("ncores", 1),
        interp_order=interp_order,
        verbose=demons_kwargs.pop("verbose", False),
        # Other kwargs (if demons_fn supports them):
        **demons_kwargs,
    )

    return result_img, out_tfm, dvf


# example: only use fixed_mask
# reg_img, tfm, dvf = demons_with_soft_mask(
#     fixed=fixed_img,
#     moving=moving_img,
#     demons_fn=demons_registration,
#     fixed_mask=fixed_mask,
#     fixed_falloff_mm=10.0,
# )

# example: use fixed_mask and moving_mask
# reg_img, tfm, dvf = demons_with_soft_mask(
#     fixed=fixed_img,
#     moving=moving_img,
#     demons_fn=demons_registration,
#     fixed_mask=fixed_mask,
#     moving_mask=moving_mask,
#     propagate_fixed_weight_to_moving=True,
#     initial_transform=initial_affine,
#     fixed_falloff_mm=8.0,
#     moving_falloff_mm=6.0,
#     regularisation_kernel_mm=1.5,
#     resolution_staging=(8, 4, 2, 1),
#     iteration_staging=(50, 40, 30, 20),
#     isotropic_resample=True,
#     smoothing_sigmas=False,
#     ncores=8,
#     verbose=True,
# )



import datetime

from pydicomrt.reg.type_transform import sitk_displacement_field_to_deformable_registration_grid

from abc import ABC, abstractmethod
from pydicom.dataset import Dataset, FileDataset, FileMetaDataset
from pydicom.sequence import Sequence
from pydicom.uid import generate_uid, ImplicitVRLittleEndian

# --------------------- Basic Registration Builder ---------------------
class BaseRegistrationBuilder(ABC):
    def __init__(self, fixed_ds_list):
        assert len(fixed_ds_list) > 0, "fixed_ds_list must be non-empty"
        self.fixed_ds_list = fixed_ds_list
        self.uid_prefix = None
        self.ref_ds = fixed_ds_list[0]
        self.registration_dataset_list = []

    def set_uid_prefix(self, uid_prefix: str):
        self.uid_prefix = uid_prefix

    @property
    @abstractmethod
    def _sop_class_uid(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def _base_file_name(self) -> str:
        raise NotImplementedError

    def _generate_file_meta(self):
        file_meta = FileMetaDataset()
        file_meta.FileMetaInformationGroupLength = 202
        file_meta.FileMetaInformationVersion = b"\x00\x01"
        file_meta.TransferSyntaxUID = ImplicitVRLittleEndian
        file_meta.MediaStorageSOPClassUID = self._sop_class_uid
        file_meta.MediaStorageSOPInstanceUID = generate_uid(prefix=self.uid_prefix)
        file_meta.ImplementationClassUID = (self.uid_prefix or "") + "1"
        file_meta.ImplementationVersionName = "pydicomRT"
        return file_meta

    def _generate_base_dataset(self):
        file_meta = self._generate_file_meta()
        ds = FileDataset(self._base_file_name, {}, file_meta=file_meta, preamble=b"\0" * 128)
        return ds

    def _add_required_elements(self, ds: Dataset):
        dt = datetime.datetime.now()
        ds.SpecificCharacterSet = "ISO_IR 192"
        ds.InstanceCreationDate = dt.strftime("%Y%m%d")
        ds.InstanceCreationTime = dt.strftime("%H%M%S")
        ds.Modality = "REG"
        ds.Manufacturer = ""
        ds.ManufacturerModelName = ""
        ds.InstitutionName = ""
        ds.is_little_endian = True
        ds.is_implicit_VR = True
        ds.SOPClassUID = ds.file_meta.MediaStorageSOPClassUID
        ds.SOPInstanceUID = ds.file_meta.MediaStorageSOPInstanceUID

        return ds

    def _add_patient_information_from_ref_ds(self, ds: Dataset, ref_ds: Dataset):
        ds.PatientID = getattr(ref_ds, "PatientID", "")
        ds.PatientName = getattr(ref_ds, "PatientName", "")
        ds.PatientBirthDate = getattr(ref_ds, "PatientBirthDate", "")
        ds.PatientBirthTime = getattr(ref_ds, "PatientBirthTime", "")
        ds.PatientSex = getattr(ref_ds, "PatientSex", "")
        ds.PatientAge = getattr(ref_ds, "PatientAge", "")
        ds.PatientWeight = getattr(ref_ds, "PatientWeight", "")
        ds.PatientPosition = getattr(ref_ds, "PatientPosition", "")
        return ds

    def _add_study_information_from_ref_ds(self, ds: Dataset, ref_ds: Dataset):
        ds.StudyInstanceUID = getattr(ref_ds, "StudyInstanceUID", "")
        ds.StudyID = getattr(ref_ds, "StudyID", "")
        ds.StudyDescription = getattr(ref_ds, "StudyDescription", "")
        ds.StudyDate = getattr(ref_ds, "StudyDate", "")
        ds.StudyTime = getattr(ref_ds, "StudyTime", "")
        ds.AccessionNumber = getattr(ref_ds, "AccessionNumber", "")
        ds.ReferringPhysicianName = getattr(ref_ds, "ReferringPhysicianName", "")
        return ds

    def _add_series_information(
        self,
        ds: Dataset,
        ref_ds: Dataset,
        series_number: str = "50",
        series_desc_prefix: str = "Spatial Registration"
        ):
        dt = datetime.datetime.now()
        ds.SeriesInstanceUID = generate_uid(prefix=self.uid_prefix)
        ds.SeriesNumber = series_number
        ds.SeriesDescription = f"{series_desc_prefix} {dt.strftime('%Y%m%d%H%M')}"
        ds.SeriesDate = dt.strftime("%Y%m%d")
        ds.SeriesTime = dt.strftime("%H%M%S")
        ds.ContentLabel = "SPATIAL REGISTRATION"
        ds.ContentDescription = "Spatial Registration"
        ds.ContentCreatorName = "pydicomRT"
        ds.FrameOfReferenceUID = getattr(ref_ds, "FrameOfReferenceUID", "")
        ds.PositionReferenceIndicator = getattr(ref_ds, "PositionReferenceIndicator", "")
        return ds


# --------------------- Spatial Registration Builder ---------------------

class SpatialRegistrationBuilder(BaseRegistrationBuilder):
    def __init__(self, fixed_ds_list):
        super().__init__(fixed_ds_list)

    @property
    def _sop_class_uid(self) -> str:
        # Spatial Registration (rigid/affine) SOP Class UID
        return "1.2.840.10008.5.1.4.1.1.66.1"

    @property
    def _base_file_name(self) -> str:
        return "spatial_registration"

    def add_rigid_registration(self, moving_ds_list, rigid_transform_matrix):
        rigid_registration_block = Dataset()
        # TODO
        # ReferencedImageSequence   (moving)
        # FrameOfReferenceUID   (moving)
        # MatrixRegistrationSequence
        #     MatrixSequence
        #         FrameOfReferenceTransformationMatrixType
        #         FrameOfReferenceTransformationMatrix    (moving to fixed)
        #     RegistrationTypeCodeSequence
        # https://dicom.nema.org/medical/dicom/2016a/output/chtml/part16/chapter_D.html#DCM_125024
        #         CodeValue     125024
        #         CodingSchemeDesignator     DCM
        #         CodingSchemeVersion       01
        #         CodeMeaning     Image Content-based Alignment

        frame_of_reference_uid = getattr(moving_ds_list[0], "FrameOfReferenceUID", "")
        rigid_registration_block.FrameOfReferenceUID = frame_of_reference_uid
        rigid_registration_block.ReferencedImageSequence = Sequence()
        for moving_ds in moving_ds_list:
            referenced_image = Dataset()
            referenced_image.ReferencedSOPClassUID = moving_ds.SOPClassUID
            referenced_image.ReferencedSOPInstanceUID = moving_ds.SOPInstanceUID
            rigid_registration_block.ReferencedImageSequence.append(referenced_image)

        rigid_registration_block.MatrixRegistrationSequence = Sequence()
        matrix_registration_sequence_block = Dataset()
        matrix_registration_sequence_block.MatrixSequence = Sequence()
        matrix_sequence = Sequence()
        matrix_sequence_block = Dataset()
        matrix_sequence_block.FrameOfReferenceTransformationMatrixType = "RIGID"
        matrix_sequence_block.FrameOfReferenceTransformationMatrix = rigid_transform_matrix
        matrix_sequence.append(matrix_sequence_block)
        matrix_registration_sequence_block.MatrixSequence = matrix_sequence
        rigid_registration_block.MatrixRegistrationSequence.append(matrix_registration_sequence_block)

        rigid_registration_block.RegistrationTypeCodeSequence = Sequence()
        registration_type_code = Dataset()
        registration_type_code.CodeValue = "125024"
        registration_type_code.CodingSchemeDesignator = "DCM"
        registration_type_code.CodingSchemeVersion = "01"
        registration_type_code.CodeMeaning = "Image Content-based Alignment"
        rigid_registration_block.RegistrationTypeCodeSequence.append(registration_type_code)

        self.registration_dataset_list.append(rigid_registration_block)
        return rigid_registration_block

    def build(self):
        ds = self._generate_base_dataset()
        self._add_required_elements(ds)
        self._add_patient_information_from_ref_ds(ds, self.ref_ds)
        self._add_study_information_from_ref_ds(ds, self.ref_ds)
        self._add_series_information(ds, self.ref_ds)
        ds.RegistrationSequence = Sequence()
        for reg_ds in self.registration_dataset_list:
            ds.RegistrationSequence.append(reg_ds)
        return ds


# --------------------- Deformable Spatial Registration Builder ---------------------

class DeformableSpatialRegistrationBuilder(BaseRegistrationBuilder):
    def __init__(self, fixed_ds_list):
        super().__init__(fixed_ds_list)

    @property
    def _sop_class_uid(self) -> str:
        # Deformable Spatial Registration SOP Class UID
        return "1.2.840.10008.5.1.4.1.1.66.3"

    @property
    def _base_file_name(self) -> str:
        return "deformable_spatial_registration"

    def add_deformable_registration(
        self,
        moving_ds_list,
        vectorial_field_transform,
        pre_transform,
        post_transform):
        deformable_registration_block = Dataset()

        # TODO
        # ReferencedImageSequence   (moving)
        # SourceFrameOfReferenceUID   (moving)
        # DeformableRegistrationGridSequence
        #     ImagePositionPatient
        #     ImageOrientationPatient
        #     GridDimensions
        #     GridResolution
        #     VectorGridData
        # PreDeformationMatrixRegistrationSequence
        #     FrameOfReferenceTransformationMatrixType
        #     FrameOfReferenceTransformationMatrix
        # PostDeformationMatrixRegistrationSequence
        #     FrameOfReferenceTransformationMatrixType
        #     FrameOfReferenceTransformationMatrix
        # RegistrationTypeCodeSequence
        # https://dicom.nema.org/medical/dicom/2016a/output/chtml/part16/chapter_D.html#DCM_125024
        #     CodeValue     125024
        #     CodingSchemeDesignator     DCM
        #     CodingSchemeVersion       01
        #     CodeMeaning     Image Content-based Alignment
        deformable_registration_block.ReferencedImageSequence = Sequence()
        deformable_registration_block.DeformableRegistrationGridSequence = Sequence()
        deformable_registration_block.PreDeformationMatrixRegistrationSequence = Sequence()
        deformable_registration_block.PostDeformationMatrixRegistrationSequence = Sequence()
        deformable_registration_block.RegistrationTypeCodeSequence = Sequence()
        deformable_registration_block.SourceFrameOfReferenceUID = getattr(moving_ds_list[0], "FrameOfReferenceUID", "")

        for moving_ds in moving_ds_list:
            referenced_image = Dataset()
            referenced_image.ReferencedSOPClassUID = moving_ds.SOPClassUID
            referenced_image.ReferencedSOPInstanceUID = moving_ds.SOPInstanceUID
            deformable_registration_block.ReferencedImageSequence.append(referenced_image)

        deformable_registration_grid = sitk_displacement_field_to_deformable_registration_grid(vectorial_field_transform)
        deformable_registration_block.DeformableRegistrationGridSequence.append(deformable_registration_grid)

        pre_deformation_matrix_registration = Dataset()
        pre_deformation_matrix_registration.FrameOfReferenceTransformationMatrixType = "RIGID"
        pre_deformation_matrix_registration.FrameOfReferenceTransformationMatrix = pre_transform
        deformable_registration_block.PreDeformationMatrixRegistrationSequence.append(pre_deformation_matrix_registration)

        post_deformation_matrix_registration = Dataset()
        post_deformation_matrix_registration.FrameOfReferenceTransformationMatrixType = "RIGID"
        post_deformation_matrix_registration.FrameOfReferenceTransformationMatrix = post_transform
        deformable_registration_block.PostDeformationMatrixRegistrationSequence.append(post_deformation_matrix_registration)

        registration_type_code = Dataset()
        registration_type_code.CodeValue = "125024"
        registration_type_code.CodingSchemeDesignator = "DCM"
        registration_type_code.CodingSchemeVersion = "01"
        registration_type_code.CodeMeaning = "Image Content-based Alignment"
        deformable_registration_block.RegistrationTypeCodeSequence.append(registration_type_code)

        self.registration_dataset_list.append(deformable_registration_block)
        pass

    def build(self):
        ds = self._generate_base_dataset()
        self._add_required_elements(ds)
        self._add_patient_information_from_ref_ds(ds, self.ref_ds)
        self._add_study_information_from_ref_ds(ds, self.ref_ds)
        self._add_series_information(ds, self.ref_ds, series_desc_prefix="Deformable Spatial Registration")
        ds.DeformableRegistrationSequence = Sequence()
        for reg_ds in self.registration_dataset_list:
            ds.DeformableRegistrationSequence.append(reg_ds)
        return ds


if __name__ == "__main__":
    from pydicomrt.utils import load_sorted_image_series
    fixed_dcm_path = "example/data/CT_001"
    moving_dcm_path = "example/data/CT_001"
    rigid_transform_matrix = [
        1, 0, 0, 0,
        0, 1, 0, 0,
        0, 0, 1, 0,
        0, 0, 0, 1,
    ]

    fixed_ds_list = load_sorted_image_series(fixed_dcm_path)
    moving_ds_list = load_sorted_image_series(moving_dcm_path)

    spatial_registration_builder = SpatialRegistrationBuilder(fixed_ds_list)
    spatial_registration_builder.add_rigid_registration(moving_ds_list, rigid_transform_matrix)
    ds = spatial_registration_builder.build()
    ds.save_as("./CT_001_rigid_registration.dcm")




import numpy as np
import SimpleITK as sitk

from pydicom.dataset import Dataset

def get_spatial_reg_dict(reg_ds: Dataset, calc_inverse_matrix: bool = False) -> dict:
    """
    Get the spatial registration dictionary from the registration dataset.
    Args:
        reg_ds (Dataset): The registration dataset.
        calc_inverse_matrix (bool): Whether to calculate the inverse matrix.
    Returns:
        dict: The spatial registration dictionary.

        dict: {
            a_foruid: {
                b_foruid: matrix_matrix,
                c_foruid: matrix_matrix,
                ...
            },
            b_foruid: {
                a_foruid: matrix_matrix,
                ...
            },
            c_foruid: {
                a_foruid: matrix_matrix,
                ...
            },

        }
    """
    reg_dict = {}
    fixed_frame_of_reference_uid = reg_ds.FrameOfReferenceUID
    reg_dict[fixed_frame_of_reference_uid] = {}

    for reg_index, reg in enumerate(reg_ds.RegistrationSequence):
        moving_frame_of_reference_uid = getattr(reg, "FrameOfReferenceUID", None)
        matrix_registration_sequence = getattr(reg, "MatrixRegistrationSequence", None)
        if moving_frame_of_reference_uid is None:
            raise ValueError("Moving frame of reference UID is not found.")
        if matrix_registration_sequence is None or len(matrix_registration_sequence) == 0:
            print(f"Matrix registration sequence is empty for {moving_frame_of_reference_uid}.")
            continue

        try:
            matrix_matrix = matrix_registration_sequence[-1].MatrixSequence[-1].FrameOfReferenceTransformationMatrix
        except:
            print(f"Matrix is not found for {moving_frame_of_reference_uid}.")
            continue
        reg_dict[fixed_frame_of_reference_uid][moving_frame_of_reference_uid] = matrix_matrix

        # for matrix_reg_index, matrix_registration in enumerate(matrix_registration_sequence):
        #     matrix_sequence = matrix_registration.MatrixSequence
        #     if matrix_sequence is None or len(matrix_sequence) == 0:
        #         print(f"Matrix sequence is empty for {moving_frame_of_reference_uid}.")
        #         continue
        #     for matrix_index, matrix in enumerate(matrix_sequence):
        #         matrix_type = getattr(matrix, "FrameOfReferenceTransformationMatrixType", None)
        #         matrix_matrix = getattr(matrix, "FrameOfReferenceTransformationMatrix", None)
        #         if matrix_type is None or matrix_matrix is None:
        #             raise ValueError("Matrix type or matrix is not found.")
        #         if matrix_matrix is None:
        #             raise ValueError("Matrix is not found.")
        #         reg_dict[fixed_frame_of_reference_uid][moving_frame_of_reference_uid] = matrix_matrix

    if calc_inverse_matrix:
        reversed_reg_dict = {}
        for fixed_frame_of_reference_uid, moving_frame_of_reference_uid_dict in reg_dict.items():
            for moving_frame_of_reference_uid, matrix_matrix in moving_frame_of_reference_uid_dict.items():
                matrix_matrix = np.array(matrix_matrix).reshape((4, 4))
                inverse_matrix = np.linalg.inv(matrix_matrix)
                inverse_matrix = inverse_matrix.ravel().tolist()
                reversed_reg_dict[moving_frame_of_reference_uid] = {}
                reversed_reg_dict[moving_frame_of_reference_uid][fixed_frame_of_reference_uid] = inverse_matrix
        reg_dict.update(reversed_reg_dict)
    return reg_dict


def get_deformable_reg_list(reg_ds: Dataset) -> list:
    reg_dict_list = []
    for reg in reg_ds.DeformableRegistrationSequence:
        reg_dict = {}
        reg_dict['SourceFrameOfReferenceUID'] = reg.SourceFrameOfReferenceUID
        reg_dict['PreDeformationMatrixRegistration'] = reg.PreDeformationMatrixRegistrationSequence[-1].FrameOfReferenceTransformationMatrix
        reg_dict['PostDeformationMatrixRegistration'] = reg.PostDeformationMatrixRegistrationSequence[-1].FrameOfReferenceTransformationMatrix
        reg_dict['DeformableRegistrationGrid'] = {}
        deformed_grid_data = reg.DeformableRegistrationGridSequence[-1]

        reg_dict['DeformableRegistrationGrid']['GridDimensions'] = deformed_grid_data.GridDimensions
        reg_dict['DeformableRegistrationGrid']['GridResolution'] = deformed_grid_data.GridResolution
        reg_dict['DeformableRegistrationGrid']['ImagePositionPatient'] = deformed_grid_data.ImagePositionPatient
        reg_dict['DeformableRegistrationGrid']['ImageOrientationPatient'] = deformed_grid_data.ImageOrientationPatient

        grid_dim = deformed_grid_data.GridDimensions
        vector_grid_data = deformed_grid_data.VectorGridData
        # vector_grid_data = unpack(f"<{len(vector_grid_data) // 4}f", vector_grid_data)
        # deformed_array = np.reshape(vector_grid_data, grid_dim[::-1] + [3,])

        vector_grid_unpack_data = np.frombuffer(vector_grid_data, dtype='<f4')
        deformed_array = vector_grid_unpack_data.reshape((grid_dim[2], grid_dim[1], grid_dim[0], 3))

        reg_dict['DeformableRegistrationGrid']['VectorGridData'] = deformed_array
        reg_dict_list.append(reg_dict)

    return reg_dict_list


if __name__ == '__main__':
    import pydicom
    deformable_dcm_path = 'example/data/DF_001/REG/DR.dcm'
    reg_ds = pydicom.dcmread(deformable_dcm_path)
    reg_dict_list = get_deformable_reg_list(reg_ds)
    # print(reg_dict_list)
    # print(reg_dict_list[0]['DeformableRegistrationGrid']['VectorGridData'].shape)
    # print(reg_dict_list[0]['DeformableRegistrationGrid']['GridDimensions'])
    # print(reg_dict_list[0]['DeformableRegistrationGrid']['GridResolution'])
    # print(reg_dict_list[0]['DeformableRegistrationGrid']['ImagePositionPatient'])
    # print(reg_dict_list[0]['DeformableRegistrationGrid']['ImageOrientationPatient'])

    spatial_reg_dcm_path = 'example/data/DF_001/REG/RE.dcm'
    spatial_reg_ds = pydicom.dcmread(spatial_reg_dcm_path)
    spatial_reg_dict = get_spatial_reg_dict(spatial_reg_ds, calc_inverse_matrix=True)
    print(spatial_reg_dict)


import cv2
import numpy as np

from pydicom.dataset import Dataset
from pydicom.sequence import Sequence

from pydicomrt.utils.coordinate_transform import apply_transformation_to_3d_points, get_pixel_to_patient_transformation_matrix
from .contour_process_method import contour_process

# ---------------------------------- interface ----------------------------------------- #

def add_contour_sequence_from_dcm_ctr_dict(
    rs_ds: Dataset,
    image_ds_list: list[Dataset],
    roi_number: int,
    dcm_ctr_dict: dict
    ) -> Sequence:
    if not hasattr(rs_ds, 'ROIContourSequence'):
        raise ValueError("ROIContourSequence does not exist")

    for index, roi_contour_sequence in enumerate(rs_ds.ROIContourSequence):
        roi_contour_sequence.get("ReferencedROINumber", None)
        if roi_contour_sequence.ReferencedROINumber == roi_number:
            roi_contour_sequence.ContourSequence = create_contour_sequence_from_dcm_ctr_dict(
                image_ds_list,
                dcm_ctr_dict,
            )
            rs_ds.ROIContourSequence[index] = roi_contour_sequence
            return rs_ds
        else:
            continue
    raise ValueError("roi_number not found")

def add_contour_sequence_from_mask3d(
    rs_ds: Dataset,
    image_ds_list: list[Dataset],
    roi_number: int,
    mask_volume: np.ndarray,
    ctr_config: dict = {
        "ex_noise_size": 10,
        "in_noise_size": 10,
        "lowpass_ratio": 10,
        "ctr_precision": 8
        }
    ) -> Sequence:
    if not hasattr(rs_ds, 'ROIContourSequence'):
        raise ValueError("ROIContourSequence does not exist")

    for index, roi_contour_sequence in enumerate(rs_ds.ROIContourSequence):
        roi_contour_sequence.get("ReferencedROINumber", None)
        if roi_contour_sequence.ReferencedROINumber == roi_number:
            roi_contour_sequence.ContourSequence = create_contour_sequence_from_mask3d(
                image_ds_list,
                mask_volume,
                ctr_config,
            )
            rs_ds.ROIContourSequence[index] = roi_contour_sequence
            return rs_ds
        else:
            continue
    raise ValueError("roi_number not found")


# ------------------------------ from_dcm_ctr_dict ------------------------------------------------- #

def create_contour_sequence_from_dcm_ctr_dict(
    image_ds_list: list[Dataset],
    dcm_ctr_dict: dict
    ) -> Sequence:
    """
    dcm_ctr_dict: {
        "sop_instance_uid": {
            "contours": list,
            "sop_class_uid": str,
        }
    }
    """
    contour_sequence = Sequence()
    for sop_instance_uid, ctr_info in dcm_ctr_dict.items():
        sop_class_uid = ctr_info["sop_class_uid"]
        ctr_list = ctr_info["contours"]
        for contour_data in ctr_list:
            contour_sequence.append(create_contour_sequence_block(contour_data, sop_class_uid, sop_instance_uid))
    return contour_sequence

def create_contour_sequence_block(contour_data, sop_class_uid, sop_instance_uid) -> Dataset:
    contour_image = Dataset()
    contour_image.ReferencedSOPClassUID = sop_class_uid
    contour_image.ReferencedSOPInstanceUID = sop_instance_uid

    contour_image_sequence = Sequence()
    contour_image_sequence.append(contour_image)

    contour = Dataset()
    contour.ContourGeometricType = "CLOSED_PLANAR"
    contour.ContourImageSequence = contour_image_sequence
    contour.NumberOfContourPoints = (len(contour_data) / 3)
    contour.ContourData = list(contour_data)

    return contour


# ------------------------------------ from_mask3d ------------------------------------------------- #

def create_contour_sequence_from_mask3d(
    image_ds_list: list[Dataset],
    mask_volume: np.ndarray,
    ctr_config: dict
    ) -> Sequence:
    contour_sequence = Sequence()
    contour_data_seq = volume_to_contour_list(mask_volume, image_ds_list, ctr_config)
    for index, [contour_data, image_slice] in enumerate(contour_data_seq):
        sop_class_uid = image_slice.SOPClassUID
        sop_instance_uid = image_slice.SOPInstanceUID
        contour = create_contour_sequence_block(contour_data, sop_class_uid, sop_instance_uid)
        contour_sequence.append(contour)

    return contour_sequence

def volume_to_contour_list(mask_volume, image_series, ctr_config):
    formatted_contours = []
    for index, series_slice in enumerate(image_series):
        mask_slice = mask_volume[index, :, :]           # mask_volume.shape = (z, y, x)     ex: (144, 512, 512)
        if np.sum(mask_slice) == 0:
            continue
        cv2_contours, hierarchy = cv2.findContours(
            mask_slice.astype(np.uint8),
            cv2.RETR_CCOMP,
            cv2.CHAIN_APPROX_NONE
            )
        if cv2_contours is None or hierarchy is None: continue
        for hier_index, contour in enumerate(cv2_contours):
            poly_hierarchy = hierarchy[0][hier_index][3]
            x_points, y_points = contour.T[:, 0]
            x_points, y_points = contour_process(x_points, y_points, poly_hierarchy,
                                                 external_noise_size=ctr_config["ex_noise_size"],
                                                 internal_noise_size=ctr_config["in_noise_size"],
                                                 low_pass_ratio=ctr_config["lowpass_ratio"])
            if len(x_points) < 3 or len(y_points) < 3: continue
            contour = np.stack([x_points, y_points]).T
            contour = np.concatenate((np.array(contour), np.full((len(contour), 1), index)), axis=1)    # add index into contour-z-axis
            transformation_matrix = get_pixel_to_patient_transformation_matrix(image_series)
            transformed_contour = apply_transformation_to_3d_points(contour, transformation_matrix)
            dicom_formatted_contour = np.ravel(transformed_contour).tolist()
            dicom_formatted_contour = [round(crood, ctr_config["ctr_precision"]) for crood in dicom_formatted_contour]
            formatted_contours.append([dicom_formatted_contour, series_slice])
    return formatted_contours



# import numpy as np
# import math

# from scipy import fft

# def contour_process(x_points, y_points, hierarchy, *args, **kwargs):
#     # print(x_points, y_points, hierarchy, args, kwargs)
#     if len(x_points) == 0 or len(y_points) == 0: return x_points, y_points  # check if contour is empty
    
#     if ("external_noise_size" in kwargs) and hierarchy == -1:
#         external_noise_size = kwargs["external_noise_size"]
#         x_points, y_points = ctr_external_denoise(x_points, y_points, noise_size=external_noise_size)
#         if len(x_points) == 0 or len(y_points) == 0: return x_points, y_points

#     if ("internal_noise_size" in kwargs) and hierarchy != -1:
#         internal_noise_size = kwargs["internal_noise_size"]
#         x_points, y_points = ctr_internal_denoise(x_points, y_points, noise_size=internal_noise_size)
#         if len(x_points) == 0 or len(y_points) == 0: return x_points, y_points
    
#     if "low_pass_ratio" in kwargs:
#         low_pass_ratio = kwargs["low_pass_ratio"]
#         x_points, y_points = ctr_fft_smooth(x_points, y_points, low_pass_ratio)

#     return x_points, y_points


# def ctr_internal_denoise(x_points, y_points, noise_size):
#     if len(x_points) < noise_size or len(y_points) < noise_size:
#         x_points = []
#         y_points = []
#     return x_points, y_points


# def ctr_external_denoise(x_points, y_points, noise_size):
#     if len(x_points) < noise_size or len(y_points) < noise_size:
#         x_points = []
#         y_points = []
#     return x_points, y_points


# def ctr_fft_smooth(x_points, y_points, low_pass_ratio=10):
#     contour_len = len(x_points)
#     if contour_len < 3:             # check if contour is too short
#         return x_points, y_points
    
#     x_points, y_points = resample_2d_contour(x_points, y_points)
#     x_points, y_points = fft_low_pass(x_points, y_points, filter_ratio=low_pass_ratio)
#     x_points = x_points[:contour_len]
#     y_points = y_points[:contour_len]
#     return x_points, y_points


# def resample_2d_contour(x, y, min_pt=1024):
#     repeat = math.ceil(2 + (min_pt / len(x)))
#     new_x = np.tile(x, repeat)
#     new_y = np.tile(y, repeat)
#     return new_x, new_y


# def fft_low_pass(x, y, filter_ratio=8):
#     """Apply low pass filter to x and y arrays by zeroing out high freq components.
    
#     ** actually, it's a band pass filter **

#     Parameters
#     ----------
#     x, y : array_like
#         Input arrays.
#     length : int > len(x)
#         The pass frequency length in frequency domain

#     Returns
#     -------
#     np.real(newZ), np.imag(newZ) : array_like
#         Filtered arrays of x and y
#     """
#     if filter_ratio >= 45:
#         return x, y

#     z = x + 1j * y    # Let x, y info together in signal space
#     C = fft.fft(z)  # R -> C

#     demod = np.abs(C)
#     indices = np.where(demod > 10)[0]

#     wave_length = len(x)
#     pass_length = int(wave_length * filter_ratio * 0.01)
#     mask = np.zeros_like(C, dtype=float)
#     length = pass_length + 1

#     # fix the bug if filtered signal's main freq only 0
#     try:
#         if length < indices[1]:
#             length = indices[1] + 1
#     except IndexError:
#         print("no other freq. > 10")
#         length = int(wave_length * 10 * 0.01)
    
#     mask[0] = 1
#     mask[1:length] = 1
#     mask[-length:] = 1

#     C *= mask   # apply filter mask
    
#     newZ = fft.ifft(C)  # C -> R
#     return np.real(newZ), np.imag(newZ)

import numpy as np
import math

def contour_process(x_points, y_points, hierarchy, *args, **kwargs):
    # print(x_points, y_points, hierarchy, args, kwargs)
    if len(x_points) == 0 or len(y_points) == 0:
        return x_points, y_points  # check if contour is empty
    
    if ("external_noise_size" in kwargs) and hierarchy == -1:
        external_noise_size = kwargs["external_noise_size"]
        x_points, y_points = ctr_external_denoise(x_points, y_points, noise_size=external_noise_size)
        if len(x_points) == 0 or len(y_points) == 0:
            return x_points, y_points

    if ("internal_noise_size" in kwargs) and hierarchy != -1:
        internal_noise_size = kwargs["internal_noise_size"]
        x_points, y_points = ctr_internal_denoise(x_points, y_points, noise_size=internal_noise_size)
        if len(x_points) == 0 or len(y_points) == 0:
            return x_points, y_points
    
    if "low_pass_ratio" in kwargs:
        low_pass_ratio = kwargs["low_pass_ratio"]
        x_points, y_points = ctr_fft_smooth(x_points, y_points, low_pass_ratio)

    return x_points, y_points


def ctr_internal_denoise(x_points, y_points, noise_size):
    if len(x_points) < noise_size or len(y_points) < noise_size:
        x_points = []
        y_points = []
    return x_points, y_points


def ctr_external_denoise(x_points, y_points, noise_size):
    if len(x_points) < noise_size or len(y_points) < noise_size:
        x_points = []
        y_points = []
    return x_points, y_points


def ctr_fft_smooth(x_points, y_points, low_pass_ratio=10):
    contour_len = len(x_points)
    if contour_len < 3:
        return x_points, y_points  # check if contour is too short
    
    x_points, y_points = resample_2d_contour(x_points, y_points)
    x_points, y_points = fft_low_pass(x_points, y_points, filter_ratio=low_pass_ratio)
    x_points = x_points[:contour_len]
    y_points = y_points[:contour_len]
    return x_points, y_points


def resample_2d_contour(x, y, min_pt=1024):
    repeat = math.ceil(2 + (min_pt / len(x)))
    new_x = np.tile(x, repeat)
    new_y = np.tile(y, repeat)
    return new_x, new_y


def fft_low_pass(x, y, filter_ratio=8):
    """Apply low pass filter to x and y arrays by zeroing out high freq components.
    
    ** actually, it's a band pass filter **

    Parameters
    ----------
    x, y : array_like
        Input arrays.
    length : int > len(x)
        The pass frequency length in frequency domain

    Returns
    -------
    np.real(newZ), np.imag(newZ) : array_like
        Filtered arrays of x and y
    """
    if filter_ratio >= 45:
        return x, y

    # Combine x and y into a complex signal
    z = x + 1j * y
    # Compute FFT using numpy
    C = np.fft.fft(z)

    demod = np.abs(C)
    indices = np.where(demod > 10)[0]

    wave_length = len(x)
    pass_length = int(wave_length * filter_ratio * 0.01)
    mask = np.zeros_like(C, dtype=float)
    length = pass_length + 1

    # fix the bug if filtered signal's main freq only 0
    try:
        if length < indices[1]:
            length = indices[1] + 1
    except IndexError:
        print("no other freq. > 10")
        length = int(wave_length * 10 * 0.01)
    
    mask[0] = 1
    mask[1:length] = 1
    mask[-length:] = 1

    # Apply low-pass mask
    C *= mask
    
    # Compute inverse FFT using numpy
    newZ = np.fft.ifft(C)
    return np.real(newZ), np.imag(newZ)



'''
Author:   BearLu
Status:   Final
Created:  2024-10-11
Python-Version: 3.10.0
Coding: UTF-8
'''

import numpy as np
import cv2


def self_round(x: float, decimal: int) -> float:
    x = x * 10**decimal
    if x >= 0:
        return float(int(x + 0.5)) / 10**decimal
    else:
        return float(int(x - 0.5)) / 10**decimal


def get_normal_vector(contour_data):
    ctrs = np.reshape(contour_data, (-1, 3))
    normal_list = [np.cross(ctrs[i + 2] - ctrs[0], ctrs[i + 1] - ctrs[0]) for i in range(0, len(ctrs) - 2, 2)]
    normal_list = [normal / np.linalg.norm(normal) for normal in normal_list]
    normal_vector = np.mean(normal_list, axis=0)
    normal_vector = normal_vector / np.linalg.norm(normal_vector)
    return normal_vector


def get_normal_vector_bear(contour_data):
    ctrs = np.reshape(contour_data, (-1, 3))
    min_index = np.unravel_index(np.argmin(ctrs[:, 0] + ctrs[:, 1]), ctrs[:, 0].shape)
    x_max_index = np.unravel_index(np.argmax(ctrs[:, 0]), ctrs[:, 0].shape)
    y_max_index = np.unravel_index(np.argmax(ctrs[:, 1]), ctrs[:, 1].shape)

    v1 = ctrs[min_index] - ctrs[x_max_index]
    v2 = ctrs[min_index] - ctrs[y_max_index]
    normal_vector = np.cross(v1, v2)
    normal_vector = normal_vector / np.linalg.norm(normal_vector)
    return normal_vector


def get_slice_index(contour_data, normal_vector):
    ctrs = np.reshape(contour_data, (-1, 3))
    index = np.dot(ctrs[0], normal_vector)
    return index


def project_point_onto_plane(point, plane_normal, d):
    '''
    a, b, c = plane_normal
    x0, y0, z0 = origin
    d = a * x0 + b * y0 + c * z0
    '''
    x_p, y_p, z_p = point
    a, b, c = plane_normal
    D = (a * x_p + b * y_p + c * z_p + d) / np.sqrt(a**2 + b**2 + c**2)
    proj_vector = np.array([a * D, b * D, c * D]) / np.sqrt(a**2 + b**2 + c**2)
    projected_point = np.array([x_p, y_p, z_p]) - proj_vector
    return projected_point


def calc_normal_projection(points: np.ndarray, normal_vector: np.ndarray) -> np.ndarray:
    """
    Calculate normal projection for a set of 3D points.
    Args:
        points: 3D points in homogeneous coordinates. (n*3 array)
        normal_vector: Normal vector of the plane.  (1*3 array)
    Returns:
        Normal projection of the points in the plane.    (n*1 array)
    """
    projection_list = [np.dot(point, normal_vector) for point in points]
    return projection_list


def calc_obb(points: np.ndarray, normal_vector: np.ndarray) -> np.ndarray:
    """
    Calculate oriented bounding box for a set of 3D points.
    Args:
        points: 3D points in homogeneous coordinates. (n*3 array)
        normal_vector: Normal vector of the plane.  (1*3 array)
    Returns:
        Oriented bounding box in homogeneous coordinates.    (8*3 array)
    """
    v1 = np.cross([0, 1, 0], normal_vector)
    if np.linalg.norm(v1) == 0:
        v1 = np.cross([1, 0, 0], normal_vector)
    v1 = v1 / np.linalg.norm(v1)
    v2 = np.cross(normal_vector, v1)
    v2 = v2 / np.linalg.norm(v2)
    R = np.array([v1, v2, normal_vector]).T
    points_rotated = points @ R
    min_bound = np.min(points_rotated, axis=0)
    max_bound = np.max(points_rotated, axis=0)
    obb_corners_local = np.array([[min_bound[0], min_bound[1], min_bound[2]],   # origin
                                  [min_bound[0], min_bound[1], max_bound[2]],   # z
                                  [min_bound[0], max_bound[1], min_bound[2]],   # y
                                  [max_bound[0], min_bound[1], min_bound[2]],   # x
                                  [min_bound[0], max_bound[1], max_bound[2]],   # yz
                                  [max_bound[0], min_bound[1], max_bound[2]],   # zx
                                  [max_bound[0], max_bound[1], min_bound[2]],   # xy
                                  [max_bound[0], max_bound[1], max_bound[2]]])  # xyz
    obb_corners_original = obb_corners_local @ R.T
    obb_corners_original
    return obb_corners_original


def apply_transformation_to_3d_points(points: np.ndarray, transformation_matrix: np.ndarray):
    vec = np.concatenate((points, np.ones((points.shape[0], 1))), axis=1)
    return vec.dot(transformation_matrix.T)[:, :3]


def round_3d_points(points):
    return (points + 0.5).astype(int)


def get_roi_number_name_map(rs_ds):
    roi_number_name_map = {}
    if hasattr(rs_ds, 'StructureSetROISequence'):
        for ssroi in rs_ds.StructureSetROISequence:
            roi_name = ssroi.ROIName
            roi_number = ssroi.ROINumber
            roi_number_name_map[roi_number] = roi_name
    return roi_number_name_map


def get_series_z_spacing(series_ds_list):
    slice_position_list = [ds.ImagePositionPatient for ds in series_ds_list]
    slice_distance_list = [np.linalg.norm(np.array(slice_position_list[index + 1]) - np.array(slice_position_list[index])) for index in range(len(slice_position_list) - 1)]
    slice_distance_list = [dist for dist in slice_distance_list if dist > 0.1]
    if not slice_distance_list:
        raise Exception("No valid slice distance found")
    return min(slice_distance_list)


def calc_image_series_affine_mapping(series_ds_list):
    first_slice = series_ds_list[0]
    origin = np.array(first_slice.ImagePositionPatient)
    x_spacing, y_spacing = first_slice.PixelSpacing
    z_spacing = get_series_z_spacing(series_ds_list)
    v_x = np.array(first_slice.ImageOrientationPatient[:3])
    v_y = np.array(first_slice.ImageOrientationPatient[3:])
    v_z = np.cross(v_x, v_y)
    # check image orientation is legal
    if not np.allclose(np.dot(v_x, v_y), 0.0, atol=1e-3) or not np.allclose(np.linalg.norm(v_z), 1.0, atol=1e-3):
        raise Exception("Invalid Image Orientation (Patient) attribute")
    linear = np.identity(3, dtype=np.float32)
    linear[0, :3] = v_x / x_spacing
    linear[1, :3] = v_y / y_spacing
    linear[2, :3] = v_z / z_spacing
    affine_mapping = np.identity(4, dtype=np.float32)
    affine_mapping[:3, :3] = linear
    affine_mapping[:3, 3] = origin.dot(-linear.T)
    mask_volume_shape = (len(series_ds_list), first_slice.Rows, first_slice.Columns)
    return affine_mapping, mask_volume_shape


def calc_rs_affine_mapping(rs_ds):
    if hasattr(rs_ds, 'ROIContourSequence'):
        all_slice_ctr = {}
        all_ctr = []
        all_normal_vector = []
        for roi_contour in rs_ds.ROIContourSequence:
            if hasattr(roi_contour, 'ContourSequence'):
                for rtroi in roi_contour.ContourSequence:
                    poly = rtroi.ContourData
                    all_ctr.extend(poly)
                    if len(poly) < 9: continue
                    # v = get_normal_vector_bear(poly)
                    normal = get_normal_vector(poly)
                    all_normal_vector.append(normal)
                    sop_instance_uid = rtroi.ContourImageSequence[0].ReferencedSOPInstanceUID   # careful use
                    if sop_instance_uid not in all_slice_ctr:
                        all_slice_ctr[sop_instance_uid] = [poly]
                    else:
                        all_slice_ctr[sop_instance_uid].append(poly)

        ##### calc normal vector #####
        all_normal_vector = np.array(all_normal_vector)
        mean_normal = np.mean(all_normal_vector, axis=0)
        normal = mean_normal / np.linalg.norm(mean_normal)
        # print(normal)
        all_ctr = np.reshape(all_ctr, (-1, 3))
        # print(all_ctr.shape)
        # print(all_ctr[:, 0].min(), all_ctr[:, 0].max())
        # print(all_ctr[:, 1].min(), all_ctr[:, 1].max())
        # print(all_ctr[:, 2].min(), all_ctr[:, 2].max())
        # print(all_ctr)
        ##### calc orented bounding box #####
        obb = calc_obb(all_ctr, normal)
        print(obb, obb.shape)
        origin = obb[0]
        # xyz = obb[7]
        v_z = obb[1] - obb[0]
        v_y = obb[2] - obb[0]
        v_x = obb[3] - obb[0]
        v_z = v_z / np.linalg.norm(v_z)
        v_y = v_y / np.linalg.norm(v_y)
        v_x = v_x / np.linalg.norm(v_x)
        # print(v_z, v_y, v_x)
        ##### calc z_spacing #####
        projection_list = calc_normal_projection(all_ctr, normal)
        # print(projection_list)
        projection_list, index_list = np.unique(projection_list, return_index=True)
        sorted_index = np.argsort(projection_list)
        projection_list = projection_list[sorted_index]
        # print(projection_list)
        slice_distance_list = [self_round(abs(projection_list[i + 1] - projection_list[i]), 2) for i in range(len(projection_list) - 1)]
        # print(slice_distance_list)
        slice_distance_list = [x for x in slice_distance_list if x != 0]
        if len(slice_distance_list) == 0:
            raise ValueError('no slice distance list')
        # print(slice_distance_list)
        slice_distance_list = list(dict.fromkeys(slice_distance_list))
        # print(slice_distance_list)
        # print(np.min(slice_distance_list), np.max(slice_distance_list), np.median(slice_distance_list))
        ##### calc affine matrix #####
        x_spacing = 0.5
        y_spacing = 0.5
        z_spacing = min(slice_distance_list)
        linear = np.identity(3, dtype=np.float32)
        linear[0, :3] = v_x / x_spacing
        linear[1, :3] = v_y / y_spacing
        linear[2, :3] = v_z / z_spacing
        affine_mapping = np.identity(4, dtype=np.float32)
        affine_mapping[:3, :3] = linear
        affine_mapping[:3, 3] = origin.dot(-linear.T)
        # print(affine_mapping)
        new_points = apply_transformation_to_3d_points(all_ctr, affine_mapping)
        new_points = round_3d_points(new_points)
        # print(new_points[:, 0].min(), new_points[:, 0].max())
        # print(new_points[:, 1].min(), new_points[:, 1].max())
        # print(new_points[:, 2].min(), new_points[:, 2].max())
        # for point in new_points:
        #     print(point)
        mask_volume_shape = (new_points[:, 2].max() + 1 - new_points[:, 2].min(),
                             new_points[:, 1].max() + 1 - new_points[:, 1].min(),
                             new_points[:, 0].max() + 1 - new_points[:, 0].min())
    return affine_mapping, mask_volume_shape


def inverse_affine_mapping(matrix):
    # Extract the linear part (top-left 3x3 submatrix)
    linear_part = matrix[:3, :3]
    # Extract the translation vector (top-right 3x1 vector)
    translation = matrix[:3, 3]

    # Compute the inverse of the linear part
    linear_part_inv = np.linalg.inv(linear_part)

    # Compute the new translation vector
    translation_inv = -np.dot(linear_part_inv, translation)

    # Create the inverse affine matrix
    inverse_matrix = np.eye(4)  # Start with identity matrix
    inverse_matrix[:3, :3] = linear_part_inv  # Set the inverse linear part
    inverse_matrix[:3, 3] = translation_inv  # Set the inverse translation

    return inverse_matrix


def rtstruct_to_mask_dict(rs_ds, affine_mapping, mask_volume_shape, roi_list=["all"], packbits=False):
    """
    Convert a DICOM RT Structure Set to a dictionary of masks.

    Parameters:
    - rs_ds: DICOM RT Structure Set dataset.
    - affine_mapping: Affine mapping matrix.
    - mask_volume_shape: Shape of the mask volume.
    - roi_list: List of ROI names to include in the dictionary. If "all", include all ROIs.
    - packbits: Whether to pack the masks using packbits.

    Returns:
    - roi_dict: Dictionary of masks, where keys are ROI names and values are masks.
    """
    roi_number_name_map = get_roi_number_name_map(rs_ds)
    if hasattr(rs_ds, 'ROIContourSequence'):
        roi_dict = {}
        for roi_contour in rs_ds.ROIContourSequence:
            roi_number = roi_contour.ReferencedROINumber
            roi_name = roi_number_name_map[roi_number]
            if roi_list != ["all"] and roi_name not in roi_list:
                continue
            roi_dict[roi_name] = {}
            if hasattr(roi_contour, 'ContourSequence'):
                mask_volume = np.zeros(mask_volume_shape, dtype=np.int32)
                for rtroi in roi_contour.ContourSequence:
                    # sop_instance_uid = rtroi.ContourImageSequence[0].ReferencedSOPInstanceUID
                    poly = rtroi.ContourData
                    if len(poly) < 9: continue
                    ctrs = np.reshape(poly, (-1, 3))
                    transform_ctrs = apply_transformation_to_3d_points(ctrs, affine_mapping)
                    mask_index = round_3d_points(transform_ctrs)
                    mask_2d_index = mask_index[:, :2].astype(np.int32)
                    mask_slice_index = mask_index[0, 2]     # TODO check index are same
                    # print(mask_slice_index)
                    slice_mask = mask_volume[mask_slice_index, :, :].astype(np.int32)
                    tmp_mask = np.zeros_like(slice_mask, dtype=np.int32)
                    # print(slice_mask.shape)
                    # print(mask_2d_index)
                    tmp_mask = cv2.fillPoly(img=tmp_mask, pts=[mask_2d_index], color=1)
                    # print(tmp_mask.shape, tmp_mask.dtype, tmp_mask.sum(), tmp_mask.max(), tmp_mask.min())
                    slice_mask = slice_mask + tmp_mask
                    mask_volume[mask_slice_index, :, :] = slice_mask
                    # print(slice_mask.shape, slice_mask.dtype, slice_mask.sum())
                mask_volume[mask_volume % 2 == 0] = 0
                mask_volume[mask_volume % 2 == 1] = 1
                # print(mask_volume.shape, mask_volume.dtype, mask_volume.sum())
                if packbits:
                    mask_volume = np.packbits(mask_volume, axis=-1)
                roi_dict[roi_name]['mask_volume'] = mask_volume
                roi_dict[roi_name]['affine_mapping'] = affine_mapping

    return roi_dict


# if __name__ == '__main__':
    # start = time.time()

    # from rs_to_volume import load_sorted_image_series

    # mr_path = 'data/mr_data/MR'
    # # rs_path = 'data/ct_data/RS/107.dcm'
    # rs_path = 'data/mr_data/RS/920.dcm'

    # ds_list = load_sorted_image_series(mr_path)
    # rs_ds = pydicom.dcmread(rs_path)
    # # print(rs_ds[0x30060050])

    # # af_map, mask_shape = calc_image_series_affine_mapping(ds_list)
    # af_map, mask_shape = calc_rs_affine_mapping(rs_ds)
    # roi_dict = rtstruct_to_mask_dict(rs_ds, af_map, mask_shape)
    # inv_af_map = inverse_affine_mapping(af_map)
    # print(af_map, inv_af_map)

    # import random
    # from rs_builder_v2 import create_rtstruct_dataset, edit_required_elements
    # from rs_builder_v2 import add_mask3d_into_rsds
    # rs_ds = create_rtstruct_dataset(ds_list)
    # rs_ds = edit_required_elements(rs_ds, structure_set_label="RSlabel")
    # for index, roi_name in enumerate(roi_dict.keys()):
    #     roi_number = str(index + 1)
    #     roi_color = [random.randint(0, 255) for _ in range(3)]
    #     mask_volume = roi_dict[roi_name]['mask_volume']
    #     roi_description = roi_name
    #     rs_ds = add_mask3d_into_rsds(rs_ds, mask_volume, inv_af_map, roi_color, roi_number, roi_name, roi_description)

    # rs_ds.save_as('data/mr_data/RS/rs.dcm')

    # from matplotlib import pyplot as plt

    # for roi_name in roi_dict.keys():
    #     print(roi_name)
    #     if roi_name != 'MR_OpticNrv_R': continue
    #     mask_volume = roi_dict[roi_name]['mask_volume']
    #     affine_mapping = roi_dict[roi_name]['affine_mapping']
    #     for index in range(0, mask_volume.shape[0], 1):
    #         # plt.imshow(mask_volume[index, :, :], cmap='gray')
    #         # plt.show()
    #         pass

    # print(time.time() - start)



from typing import Optional, Sequence, Tuple, Union, List
import numpy as np
import SimpleITK as sitk


def smooth_and_resample(
    image,
    isotropic_voxel_size_mm=None,
    shrink_factor=None,
    smoothing_sigma=None,
    interpolator=sitk.sitkLinear,
):
    """
    Smooth (optional Gaussian) and resample an image to a lower resolution or
    isotropic voxel size.

    This function is typically used to build an image pyramid for multi-resolution
    registration. It first applies Gaussian smoothing (if requested), then resamples
    the image using either:
      - a shrink factor (downsampling), or
      - a target isotropic voxel size.

    Parameters
    ----------
    image : sitk.Image
        Input image to smooth and resample.
    isotropic_voxel_size_mm : float, optional
        Desired isotropic voxel size (mm). If provided, overrides `shrink_factor`.
    shrink_factor : float or list of floats, optional
        Downsampling factor(s). If scalar, applied equally to all dimensions.
        If list/tuple, must match the number of image dimensions.
        Mutually exclusive with `isotropic_voxel_size_mm`.
    smoothing_sigma : float or list of floats, optional
        Gaussian smoothing sigma(s), in physical units (mm). If scalar, same
        sigma is applied in all dimensions; if sequence, must match the number
        of dimensions.
    interpolator : int, default = sitk.sitkLinear
        Interpolator enum used by SimpleITK's Resample function.

    Returns
    -------
    sitk.Image
        Smoothed and resampled image.

    Raises
    ------
    AttributeError
        If both `isotropic_voxel_size_mm` and `shrink_factor` are specified.
    """

    # ---- Step 1. Optional Gaussian smoothing ----
    if smoothing_sigma:
        if hasattr(smoothing_sigma, "__iter__"):
            # Variance = sigma^2 per dimension
            smoothing_variance = [i * i for i in smoothing_sigma]
        else:
            smoothing_variance = (smoothing_sigma ** 2,) * image.GetDimension()

        # Kernel width = ~8*sigma/spacing (rounded)
        maximum_kernel_width = int(
            max([8 * j * i for i, j in zip(image.GetSpacing(), smoothing_variance)])
        )

        # Apply smoothing in physical space
        image = sitk.DiscreteGaussian(image, smoothing_variance, maximum_kernel_width)

    # ---- Step 2. Retrieve current metadata ----
    original_spacing = image.GetSpacing()
    original_size = image.GetSize()

    # ---- Step 3. Check for conflicting arguments ----
    if shrink_factor and isotropic_voxel_size_mm:
        raise AttributeError(
            "Function must be called with either isotropic_voxel_size_mm or "
            "shrink_factor, not both."
        )

    # ---- Step 4. Compute new size ----
    if isotropic_voxel_size_mm:
        # Target isotropic resolution → scale factor = target/original spacing
        scale_factor = (
            isotropic_voxel_size_mm * np.ones_like(image.GetSize()) / np.array(image.GetSpacing())
        )
        # Compute new size = old_size / scale_factor (rounded)
        new_size = [int(sz / float(sf) + 0.5) for sz, sf in zip(original_size, scale_factor)]

    elif shrink_factor:
        if isinstance(shrink_factor, (list, tuple)):
            # Per-dimension shrink factor
            new_size = [int(sz / float(sf) + 0.5) for sz, sf in zip(original_size, shrink_factor)]
        else:
            # Same shrink factor in all dimensions
            new_size = [int(sz / float(shrink_factor) + 0.5) for sz in original_size]

    else:
        # Neither shrink nor isotropic resampling → return unchanged
        return image

    # ---- Step 5. Compute new spacing from new size ----
    # Keep same physical extent → spacing = (extent / (new_size-1))
    new_spacing = [
        ((size_o_i - 1) * spacing_o_i) / (size_n_i - 1)
        for size_o_i, spacing_o_i, size_n_i in zip(original_size, original_spacing, new_size)
    ]

    # ---- Step 6. Resample with new size/spacing ----
    return sitk.Resample(
        image,
        new_size,
        sitk.Transform(),          # identity transform
        interpolator,              # chosen interpolator
        image.GetOrigin(),         # preserve origin
        new_spacing,               # computed spacing
        image.GetDirection(),      # preserve direction cosines
        0.0,                       # default background value
        image.GetPixelID(),        # preserve pixel type
    )


def multiscale_demons(
    registration_algorithm,
    fixed_image,
    moving_image,
    initial_transform=None,
    initial_displacement_field=None,
    isotropic_resample=None,
    resolution_staging=None,
    smoothing_sigmas=None,
    iteration_staging=None,
    interp_order=sitk.sitkLinear,
):
    """
    Perform deformable image registration using a multi-scale (pyramid) demons strategy.

    This function applies a given registration algorithm (e.g., Demons) in a coarse-to-fine
    hierarchical fashion. At each pyramid level, the fixed and moving images are smoothed and
    resampled, and the registration is run iteratively. The displacement vector field (DVF)
    is refined across scales until the final resolution is reached.

    Parameters
    ----------
    registration_algorithm : SimpleITK registration filter
        A demons-based registration algorithm object with
        `Execute(fixed_image, moving_image) -> displacement_field` interface.
    fixed_image : sitk.Image
        The reference image. The resulting transform maps points from this image
        to the moving image's spatial domain.
    moving_image : sitk.Image
        The moving image. The algorithm estimates a transformation from
        fixed_image → moving_image space.
    initial_transform : Optional[sitk.Transform], default = None
        Initial transform to initialize the displacement field. Ignored if
        `initial_displacement_field` is provided.
    initial_displacement_field : Optional[sitk.Image], default = None
        Optional displacement field to initialize registration. If provided,
        it overrides `initial_transform`.
    isotropic_resample : bool, default = None
        If True, interpret `resolution_staging` as isotropic voxel sizes (mm).
        If False, interpret as shrink factors (relative downsampling).
    resolution_staging : Sequence[int or float], default = None
        Per-level downsampling factors or target isotropic resolutions (mm),
        depending on `isotropic_resample`.
    smoothing_sigmas : Sequence[float], default = None
        Per-level Gaussian smoothing sigmas (in physical units). Length must
        match `resolution_staging`.
    iteration_staging : Sequence[int], default = None
        Number of iterations to run at each pyramid level.
    interp_order : int, default = sitk.sitkLinear
        Interpolator used for image resampling.

    Returns
    -------
    dvf_total : sitk.Image
        Final displacement vector field (DVF) in the fixed image space.

    Notes
    -----
    - Uses Gaussian smoothing at each scale to stabilize registration.
    - The final DVF is smoothed at each iteration and accumulated across levels.
    - Pixel type of DVF is constrained to sitkVectorFloat64 for compatibility
      with Demons filters in SimpleITK.
    """
    # Lists to hold multi-resolution fixed and moving images
    fixed_images = []
    moving_images = []

    # Construct the image pyramid: apply smoothing + resampling per scale
    for resolution, smoothing_sigma in zip(resolution_staging, smoothing_sigmas):
        isotropic_voxel_size_mm = None
        shrink_factor = None

        # Choose between isotropic voxel resampling or shrink factor
        if isotropic_resample:
            isotropic_voxel_size_mm = resolution
        else:
            shrink_factor = resolution

        # Resample fixed image
        fixed_images.append(
            smooth_and_resample(
                fixed_image,
                isotropic_voxel_size_mm=isotropic_voxel_size_mm,
                shrink_factor=shrink_factor,
                smoothing_sigma=smoothing_sigma,
                interpolator=interp_order,
            )
        )

        # Resample moving image
        moving_images.append(
            smooth_and_resample(
                moving_image,
                isotropic_voxel_size_mm=isotropic_voxel_size_mm,
                shrink_factor=shrink_factor,
                smoothing_sigma=smoothing_sigma,
                interpolator=interp_order,
            )
        )

    # Initialize displacement field
    if not initial_displacement_field:
        if initial_transform:
            # Convert transform into a displacement field image
            initial_displacement_field = sitk.TransformToDisplacementField(
                initial_transform,
                sitk.sitkVectorFloat64,
                fixed_image.GetSize(),
                fixed_image.GetOrigin(),
                fixed_image.GetSpacing(),
                fixed_image.GetDirection(),
            )
        else:
            # Create an empty displacement field image matching fixed_image size
            if len(moving_image.GetSize()) == 2:
                initial_displacement_field = sitk.Image(
                    fixed_image.GetWidth(),
                    fixed_image.GetHeight(),
                    sitk.sitkVectorFloat64,
                )
            elif len(moving_image.GetSize()) == 3:
                initial_displacement_field = sitk.Image(
                    fixed_image.GetWidth(),
                    fixed_image.GetHeight(),
                    fixed_image.GetDepth(),
                    sitk.sitkVectorFloat64,
                )
            # Copy metadata (origin, spacing, direction) from fixed_image
            initial_displacement_field.CopyInformation(fixed_image)
    else:
        # Resample given initial DVF to fixed image grid
        initial_displacement_field = sitk.Resample(initial_displacement_field, fixed_image)

    # Initialize total deformation vector field (DVF) aligned to fixed image
    dvf_total = sitk.Resample(initial_displacement_field, fixed_image)

    # Multi-scale loop: process from coarsest → finest resolution
    for i in range(len(fixed_images)):
        f_image = fixed_images[i]
        m_image = moving_images[i]

        # Resample DVF to current resolution
        dvf_total = sitk.Resample(dvf_total, f_image)

        # Convert DVF into a transform and warp the moving image
        tfm_total = sitk.DisplacementFieldTransform(sitk.Cast(dvf_total, sitk.sitkVectorFloat64))
        m_image = sitk.Resample(m_image, tfm_total, interp_order)

        # Configure number of iterations at this scale
        iters = iteration_staging[i]
        registration_algorithm.SetNumberOfIterations(iters)

        # Run demons registration at current scale
        dvf_iter = registration_algorithm.Execute(f_image, m_image)

        # Accumulate update into running DVF (compose transforms)
        dvf_total = dvf_total + sitk.Resample(dvf_iter, tfm_total)

        # Regularize/smooth the DVF
        sigma = registration_algorithm.GetStandardDeviations()
        dvf_total = sitk.SmoothingRecursiveGaussian(dvf_total, sigma)
        dvf_total = sitk.Cast(dvf_total, sitk.sitkVectorFloat64)

    # Resample final DVF to fixed image grid
    dvf_total = sitk.Resample(dvf_total, fixed_image)

    return dvf_total


def deformable_registration_command_iteration(method):
    """
    Utility function to print information during demons registration
    """
    print("{0:3} = {1:10.5f}".format(method.GetElapsedIterations(), method.GetMetric()))


def demons_registration(
    fixed_image: sitk.Image,
    moving_image: sitk.Image,
    resolution_staging: Sequence[int] = (8, 4, 1),
    iteration_staging: Sequence[int] = (10, 10, 10),
    isotropic_resample: bool = False,
    initial_displacement_field: Optional[sitk.Image] = None,
    regularisation_kernel_mm: Union[float, Sequence[float]] = 1.5,
    smoothing_sigma_factor: float = 1.0,
    smoothing_sigmas: Optional[Union[float, Sequence[float], bool]] = False,
    default_value: Optional[float] = None,
    ncores: int = 1,
    interp_order: int = sitk.sitkLinear,
    verbose: bool = False,
) -> Tuple[sitk.Image, sitk.Transform, sitk.Image]:
    """
    Deformable registration via Fast Symmetric-Forces Demons (SimpleITK) with multi-resolution strategy.

    The function registers `moving_image` to `fixed_image` by estimating a dense displacement
    field using a coarse-to-fine (image pyramid) schedule. It then applies the resulting
    displacement-field transform to resample the moving image into the fixed image space.

    Parameters
    ----------
    fixed_image : sitk.Image
        Reference (target) image. Output is defined on this image grid.
    moving_image : sitk.Image
        Image to be warped into the fixed image space.
    resolution_staging : Sequence[int], default = (8, 4, 1)
        Pyramid schedule. If `isotropic_resample=False`, these are downsampling factors.
        If `isotropic_resample=True`, these are target isotropic voxel sizes in millimeters.
        Length of this sequence defines the number of pyramid levels.
    iteration_staging : Sequence[int], default = (20, 20, 20)
        Number of iterations per pyramid level. Must match the length of `resolution_staging`.
    isotropic_resample : bool, default = False
        If True, interpret `resolution_staging` as isotropic voxel sizes (mm); otherwise as shrink factors.
    initial_displacement_field : Optional[sitk.Image], default = None
        Initial DVF (vector image). If provided, it overrides any `initial_transform` usage
        inside `multiscale_demons` and is resampled onto the fixed grid when needed.
    regularisation_kernel_mm : float or Sequence[float], default = 1.0
        Standard deviation(s) in millimeters used by demons for smoothing the update and DVF.
        Scalar broadcasts to all axes; sequence length must be 1 or image dimension.
    smoothing_sigma_factor : float, default = 1.0
        Fallback factor to derive `smoothing_sigmas` per level if `smoothing_sigmas` is not provided.
        Each level's sigma = `resolution_staging[level] * smoothing_sigma_factor`.
    smoothing_sigmas : float or Sequence[float] or bool, default = False
        Per-level Gaussian sigmas (in physical units) used prior to resampling at each level.
        - False/None: derived from `resolution_staging` and `smoothing_sigma_factor`
        - float: same sigma for all levels
        - sequence: length must match number of levels
    default_value : Optional[float], default = None
        Out-of-bounds value during final resampling. If None, use 0. For CT-like images
        (min <= -1000), default to -1000.
    ncores : int, default = 1
        Number of threads for demons filter.
    interp_order : int, default = sitk.sitkLinear
        Interpolator for resampling (e.g., sitk.sitkNearestNeighbor, sitk.sitkLinear, sitk.sitkBSpline).
    verbose : bool, default = False
        If True and a global callback `deformable_registration_command_iteration` is defined,
        prints per-iteration metric values.

    Returns
    -------
    registered_image : sitk.Image
        The moving image warped into the fixed image space.
    output_transform : sitk.Transform
        DisplacementFieldTransform built from the final DVF.
    deformation_field : sitk.Image
        The final dense displacement field (vector image) aligned to `fixed_image`.

    Raises
    ------
    ValueError
        On invalid arguments (length mismatches, non-positive values, dimension mismatch, etc.).
    RuntimeError
        If helper `multiscale_demons(...)` is not available in the current scope.

    Notes
    -----
    - Internally casts images to Float32 for demons, then casts result back to the original
      moving pixel type after resampling.
    - Regularization sigmas are specified in *voxels* to SimpleITK; this function converts
      millimeter inputs to voxel units using the fixed image spacing.
    """

    # ---- Basic validation on lists/levels/cores ----
    if not isinstance(resolution_staging, Sequence) or len(resolution_staging) == 0:
        raise ValueError("`resolution_staging` must be a non-empty sequence of positive integers.")
    if not isinstance(iteration_staging, Sequence) or len(iteration_staging) == 0:
        raise ValueError("`iteration_staging` must be a non-empty sequence of positive integers.")
    if len(iteration_staging) != len(resolution_staging):
        raise ValueError("`iteration_staging` length must match `resolution_staging` length.")
    if any(int(x) <= 0 for x in resolution_staging):
        raise ValueError("All values in `resolution_staging` must be positive integers.")
    if any(int(x) <= 0 for x in iteration_staging):
        raise ValueError("All values in `iteration_staging` must be positive integers.")
    if ncores <= 0:
        raise ValueError("`ncores` must be a positive integer.")

    # ---- Dimensionality checks ----
    dim = fixed_image.GetDimension()
    if moving_image.GetDimension() != dim:
        raise ValueError("`fixed_image` and `moving_image` must have the same dimension.")
    if initial_displacement_field is not None and initial_displacement_field.GetDimension() != dim:
        raise ValueError("`initial_displacement_field` must have the same dimension as the images.")

    # ---- Prepare per-level smoothing sigmas (in physical units) ----
    num_levels = len(resolution_staging)
    if smoothing_sigmas is False or smoothing_sigmas is None:
        # Derive from schedule × factor, e.g., (8,4,1) × 1.0 → (8.0,4.0,1.0)
        smoothing_sigmas_list: List[float] = [
            float(level) * float(smoothing_sigma_factor) for level in resolution_staging
        ]
    elif isinstance(smoothing_sigmas, (int, float)):
        smoothing_sigmas_list = [float(smoothing_sigmas)] * num_levels
    elif isinstance(smoothing_sigmas, Sequence):
        if len(smoothing_sigmas) != num_levels:
            raise ValueError("When providing a sequence, `smoothing_sigmas` length must match the number of levels.")
        smoothing_sigmas_list = [float(s) for s in smoothing_sigmas]
    else:
        raise ValueError("`smoothing_sigmas` must be False/None, a float, or a sequence of floats.")

    # ---- Convert regularisation sigma from mm → voxels for demons filter ----
    if isinstance(regularisation_kernel_mm, (int, float)):
        kernel_mm = np.full(dim, float(regularisation_kernel_mm), dtype=np.float64)
    else:
        ker = np.asarray(regularisation_kernel_mm, dtype=np.float64).ravel()
        if ker.size not in (1, dim):
            raise ValueError(
                f"`regularisation_kernel_mm` must be a float or a sequence of length 1 or {dim}."
            )
        kernel_mm = np.full(dim, float(ker[0]), dtype=np.float64) if ker.size == 1 else ker

    # Keep the original moving image pixel type to cast result back later
    moving_image_type = moving_image.GetPixelID()

    # Demons prefers float images; cast to Float32 if necessary
    if fixed_image.GetPixelID() != sitk.sitkFloat32:
        fixed_image = sitk.Cast(fixed_image, sitk.sitkFloat32)
    if moving_image.GetPixelID() != sitk.sitkFloat32:
        moving_image = sitk.Cast(moving_image, sitk.sitkFloat32)

    # ---- Configure demons filter ----
    registration_method = sitk.FastSymmetricForcesDemonsRegistrationFilter()
    registration_method.SetNumberOfThreads(int(ncores))
    registration_method.SetSmoothUpdateField(True)          # smooth incremental updates
    registration_method.SetSmoothDisplacementField(True)    # smooth accumulated DVF

    # Convert mm to voxels using fixed image spacing (what the filter expects)
    spacing = np.asarray(fixed_image.GetSpacing(), dtype=np.float64)
    if spacing.size != dim:
        raise ValueError("Unexpected spacing dimensionality mismatch.")
    kernel_vox = (kernel_mm / spacing).tolist()
    registration_method.SetStandardDeviations(kernel_vox)

    # Optional iteration callback for progress printing
    if verbose and "deformable_registration_command_iteration" in globals():
        registration_method.AddCommand(
            sitk.sitkIterationEvent,
            lambda: globals()["deformable_registration_command_iteration"](registration_method),
        )

    # ---- Run multi-scale demons to obtain the final DVF ----
    try:
        deformation_field: sitk.Image = multiscale_demons(
            registration_algorithm=registration_method,
            fixed_image=fixed_image,
            moving_image=moving_image,
            resolution_staging=tuple(int(x) for x in resolution_staging),
            smoothing_sigmas=smoothing_sigmas_list,
            iteration_staging=tuple(int(x) for x in iteration_staging),
            isotropic_resample=bool(isotropic_resample),
            initial_displacement_field=initial_displacement_field,
            interp_order=int(interp_order),
        )
    except NameError as e:
        # Provide a clearer error if helper function is missing
        raise RuntimeError(
            "Required helper `multiscale_demons(...)` is not defined in the current scope."
        ) from e

    # ---- Build a resampler to warp the moving image using the DVF ----
    resampler = sitk.ResampleImageFilter()
    resampler.SetReferenceImage(fixed_image)      # output grid = fixed image grid
    resampler.SetInterpolator(int(interp_order))  # interpolation scheme

    # Heuristic default value: use -1000 for CT-like images, else 0
    if default_value is None:
        default_value = 0.0
        try:
            arr_min = float(np.min(sitk.GetArrayViewFromImage(moving_image)))
            if arr_min <= -1000.0:
                default_value = -1000.0
        except Exception:
            # If array view is not available, keep default 0.0
            pass
    resampler.SetDefaultPixelValue(float(default_value))

    # Create a displacement-field transform from the DVF (cast to Float64 for SITK transform)
    df64 = sitk.Cast(deformation_field, sitk.sitkVectorFloat64)
    output_transform = sitk.DisplacementFieldTransform(df64)
    resampler.SetTransform(output_transform)

    # Execute resampling to get the registered (warped) image
    registered_image = resampler.Execute(moving_image)

    # Copy spatial metadata from the fixed image; cast back to the original moving pixel type
    registered_image.CopyInformation(fixed_image)
    registered_image = sitk.Cast(registered_image, moving_image_type)

    return registered_image, output_transform, deformation_field



from pydicom.dataset import Dataset
from pydicom.sequence import Sequence


def create_roi_into_rs_ds(
    rs_ds: Dataset,
    roi_color: list,
    roi_number: int,
    roi_name: str,
    roi_description: str,
    roi_interpreted_type: str = "ORGAN",
    ) -> Dataset:
    frame_of_ref_uid = rs_ds.FrameOfReferenceUID
    rs_ds.StructureSetROISequence.append(create_empty_structure_set_roi(roi_number, roi_name, roi_description, frame_of_ref_uid))
    rs_ds.ROIContourSequence.append(create_empty_roi_contour_sequence(roi_number, roi_color))
    rs_ds.RTROIObservationsSequence.append(create_empty_rtroi_observation(roi_number, roi_interpreted_type))
    return rs_ds

def create_empty_structure_set_roi(roi_number,
                                   roi_name,
                                   roi_description,
                                   frame_of_ref_uid,
                                   roi_generation_algorithm="AUTOMATIC",
                                   roi_generation_description="KumaGenerated",
                                   ) -> Dataset:
    structure_set_roi = Dataset()
    structure_set_roi.ROINumber = roi_number
    structure_set_roi.ROIName = roi_name
    structure_set_roi.ROIDescription = roi_description
    structure_set_roi.ROIGenerationAlgorithm = roi_generation_algorithm
    structure_set_roi.ROIGenerationDescription = roi_generation_description
    structure_set_roi.ReferencedFrameOfReferenceUID = frame_of_ref_uid
    return structure_set_roi

def create_empty_roi_contour_sequence(roi_number, roi_color) -> Dataset:
    roi_contour_sequence = Dataset()
    roi_contour_sequence.ReferencedROINumber = roi_number    # also in create_structure_set_roi
    roi_contour_sequence.ROIDisplayColor = roi_color
    roi_contour_sequence.ContourSequence = Sequence()
    return roi_contour_sequence

def create_empty_rtroi_observation(roi_number, roi_interpreted_type='ORGAN') -> Dataset:
    rtroi_observation = Dataset()
    rtroi_observation.ObservationNumber = roi_number
    rtroi_observation.ReferencedROINumber = roi_number
    rtroi_observation.ROIObservationDescription = "Type:Soft,Range:*/*,Fill:0,Opacity:0.0,Thickness:1,LineThickness:2,read-only:false"
    rtroi_observation.private_creators = "higumalu"
    rtroi_observation.RTROIInterpretedType = roi_interpreted_type
    rtroi_observation.ROIInterpreter = ""
    return rtroi_observation



import numpy as np
import SimpleITK as sitk
from pydicom import Dataset

def affine_to_homogeneous_matrix(transform: sitk.AffineTransform) -> np.ndarray:
    """
    Convert a SimpleITK AffineTransform to a 4x4 homogeneous transformation matrix.

    Args:
        transform (sitk.AffineTransform): The affine transform to convert.

    Returns:
        np.ndarray: 4x4 homogeneous matrix (dtype float64).
    """
    matrix3x3 = np.array(transform.GetMatrix()).reshape((3, 3))
    translation = np.array(transform.GetTranslation())
    hom_mat = np.eye(4)
    hom_mat[:3, :3] = matrix3x3
    hom_mat[:3, 3] = translation
    return hom_mat


def sitk_displacement_field_to_deformable_registration_grid(transform: sitk.DisplacementFieldTransform) -> Dataset:
    """
    Convert a SimpleITK DisplacementFieldTransform to a dictionary.
    Args:
        transform (sitk.DisplacementFieldTransform): The displacement field transform to convert.
    Returns:
        Dataset: Dataset containing deformable registration grid information.
    """
    displacement_field = sitk.GetArrayFromImage(transform.GetDisplacementField())
    displacement_field = displacement_field.astype('<f4', copy=False)
    vector_grid_data = displacement_field.ravel(order='C').tobytes()

    origin = list(map(float, transform.GetOrigin()))
    spacing = list(map(float, transform.GetSpacing()))
    size = list(map(int, transform.GetSize()))
    direction = transform.GetDirection()
    orientation = list(map(float, direction[:6]))

    deformable_registration_grid = Dataset()
    deformable_registration_grid.ImagePositionPatient = origin
    deformable_registration_grid.ImageOrientationPatient = orientation
    deformable_registration_grid.GridDimensions = size
    deformable_registration_grid.GridResolution = spacing
    deformable_registration_grid.VectorGridData = vector_grid_data

    return deformable_registration_grid



import os
import numpy as np
import SimpleITK as sitk

from datetime import datetime
from pydicom.tag import Tag
from pydicom.dataset import Dataset, FileDataset, FileMetaDataset
from pydicom.sequence import Sequence
from pydicom.uid import generate_uid
from pydicom.uid import ImplicitVRLittleEndian


DICOM_UID_PREFIX = os.getenv('DICOM_UID_PREFIX', "1.2.826.0.1.3680043.8.498.")  # PYDICOM_ROOT_UID

def generate_base_dataset() -> FileDataset:
    file_name = "bear_dose"
    file_meta = get_file_meta()
    ds = FileDataset(file_name, {}, file_meta=file_meta, preamble=b"\0" * 128)
    add_required_elements_to_ds(ds)
    add_sequence_lists_to_ds(ds)
    return ds

def get_file_meta() -> FileMetaDataset:
    file_meta = FileMetaDataset()
    file_meta.FileMetaInformationGroupLength = 190
    file_meta.FileMetaInformationVersion = b"\x00\x01"
    file_meta.TransferSyntaxUID = ImplicitVRLittleEndian
    file_meta.MediaStorageSOPClassUID = "1.2.840.10008.5.1.4.1.1.481.2"
    file_meta.MediaStorageSOPInstanceUID = (
        generate_uid(prefix=DICOM_UID_PREFIX)
    )
    file_meta.ImplementationClassUID = DICOM_UID_PREFIX + "1"
    # TODO file_meta.ImplementationVersionName = ""
    return file_meta

def add_required_elements_to_ds(ds: FileDataset):
    dt = datetime.now()
    # Append data elements required by the DICOM standarad
    ds.SpecificCharacterSet = "ISO_IR 192"
    ds.InstanceCreationDate = dt.strftime("%Y%m%d")
    ds.InstanceCreationTime = dt.strftime("%H%M%S")
    ds.Modality = "RTDOSE"
    ds.Manufacturer = "pydicomRT"
    ds.ManufacturerModelName = "modelv1"
    ds.InstitutionName = "pydicomRT"
    # Set the transfer syntax
    ds.is_little_endian = True
    ds.is_implicit_VR = True
    # Set values already defined in the file meta
    ds.SOPClassUID = ds.file_meta.MediaStorageSOPClassUID
    ds.SOPInstanceUID = ds.file_meta.MediaStorageSOPInstanceUID

    ds.ApprovalStatus = "UNAPPROVED"

def add_sequence_lists_to_ds(ds: FileDataset):
    ds.ReferencedRTPlanSequence = Sequence()

# ---------------------------------------------------------------------------------------- #

def add_patient_information(ds: FileDataset, reference_ds: Dataset):
    ds.PatientName = getattr(reference_ds, "PatientName", "Unknown")
    ds.PatientID = getattr(reference_ds, "PatientID", "Unknown")
    ds.PatientBirthDate = getattr(reference_ds, "PatientBirthDate", "")
    ds.PatientSex = getattr(reference_ds, "PatientSex", "")
    ds.PatientAge = getattr(reference_ds, "PatientAge", "")
    ds.PatientSize = getattr(reference_ds, "PatientSize", "")
    ds.PatientWeight = getattr(reference_ds, "PatientWeight", "")
    return ds

def add_study_information(ds: FileDataset, reference_ds: Dataset):
    dt = datetime.now()
    ds.StudyInstanceUID = getattr(reference_ds, "StudyInstanceUID", generate_uid(prefix=DICOM_UID_PREFIX))
    ds.StudyDate = getattr(reference_ds, "StudyDate", dt.strftime("%Y%m%d"))
    ds.StudyTime = getattr(reference_ds, "StudyTime", dt.strftime("%H%M%S"))
    ds.StudyID = getattr(reference_ds, "StudyID", "")
    ds.AccessionNumber = getattr(reference_ds, "AccessionNumber", "")
    ds.ReferringPhysicianName = getattr(reference_ds, "ReferringPhysicianName", "")
    return ds

def add_series_information(ds: FileDataset, reference_ds: Dataset, series_number: int = 1):
    dt = datetime.now()
    ds.Modality = "RTDOSE"
    ds.SeriesInstanceUID = generate_uid(prefix=DICOM_UID_PREFIX)
    ds.SeriesDate = dt.strftime("%Y%m%d")
    ds.SeriesTime = dt.strftime("%H%M%S")
    ds.SeriesDescription = getattr(reference_ds, "SeriesDescription", "") + "_dose" + dt.strftime("%Y%m%d%H%M%S")
    ds.SeriesNumber = getattr(reference_ds, "SeriesNumber", series_number)
    ds.OperatorsName = getattr(reference_ds, "OperatorsName", "")
    return ds

def add_frame_of_reference_information(ds: FileDataset, reference_ds: Dataset):
    ds.FrameOfReferenceUID = getattr(reference_ds, "FrameOfReferenceUID", generate_uid(prefix=DICOM_UID_PREFIX))
    ds.PositionReferenceIndicator = getattr(reference_ds, "PositionReferenceIndicator", "")
    return ds

def add_rf_rt_plan_seq_from_dose_ds(ds: FileDataset, reference_ds: Dataset):
    sop_class_uid = getattr(getattr(reference_ds, "file_meta", None), "MediaStorageSOPClassUID", None)
    if sop_class_uid != "1.2.840.10008.5.1.4.1.1.481.2":
        raise ValueError("reference_ds is not an RT Dose (SOPClassUID does not match RT Dose storage).")
    rf_rt_plan_seq = getattr(reference_ds, "ReferencedRTPlanSequence", None)
    if rf_rt_plan_seq is None or len(rf_rt_plan_seq) == 0:
        raise ValueError("reference_ds does not have a ReferencedRTPlanSequence.")
    ds.ReferencedRTPlanSequence = rf_rt_plan_seq
    return ds

def add_rf_rt_plan_seq_from_plan_ds(ds: FileDataset, reference_ds: Dataset):
    sop_class_uid = getattr(getattr(reference_ds, "file_meta", None), "MediaStorageSOPClassUID", None)
    if sop_class_uid != "1.2.840.10008.5.1.4.1.1.481.3":
        raise ValueError("reference_ds is not an RT Plan (SOPClassUID does not match RT Plan storage).")
    
    rf_rt_plan_seq_block = Dataset()
    rf_rt_plan_seq_block.ReferencedSOPClassUID = reference_ds.SOPClassUID
    rf_rt_plan_seq_block.ReferencedSOPInstanceUID = reference_ds.SOPInstanceUID
    # TODO ReferencedFractionGroupSequence, ReferencedPlanOverviewIndex
    '''
    (300C,0020) ReferencedFractionGroupSequence
    Required if Dose Summation Type (3004,000A) is 
    FRACTION, BEAM, BRACHY, FRACTION_SESSION, BEAM_SESSION, BRACHY_SESSION or CONTROL_POINT.

    (300C,0118) ReferencedPlanOverviewIndex
    The value of Plan Overview Index (300C,0117) from 
    the Plan Overview Sequence (300C,0116) to which this RT Plan corresponds.
    Shall be unique, i.e., not be duplicated within another Item of this Referenced RT Plan Sequence (300C,0002).
    Required if Plan Overview Sequence (300C,0116) is present.
    '''
    ds.ReferencedRTPlanSequence.append(rf_rt_plan_seq_block)
    return ds


# ---------------------------------------------------------------------------------------- #

def cp_information_from_ds(
    ds: FileDataset,
    reference_ds: Dataset,
    ):
    ds = add_patient_information(ds, reference_ds)
    ds = add_study_information(ds, reference_ds)
    ds = add_series_information(ds, reference_ds)
    ds = add_frame_of_reference_information(ds, reference_ds)
    return ds

# ---------------------------------------------------------------------------------------- #

def add_dose_grid_to_ds(
    ds: FileDataset,
    dose_sitk_image: sitk.Image,
    ):
    scaling = 1e-7
    array = sitk.GetArrayFromImage(dose_sitk_image)
    size = list(dose_sitk_image.GetSize())
    spacing = list(dose_sitk_image.GetSpacing())
    direction = list(dose_sitk_image.GetDirection())
    origin = list(dose_sitk_image.GetOrigin())

    ds.SamplesPerPixel = 1
    ds.PhotometricInterpretation = "MONOCHROME2"
    ds.BitsAllocated = 32
    ds.BitsStored = 32
    ds.HighBit = 31
    ds.PixelRepresentation = 1
    ds.DoseUnits = "GY"
    ds.DoseType = "EFFECTIVE"
    ds.DoseSummationType = "PLAN"
    ds.NumberOfFrames = size[2]
    ds.FrameIncrementPointer = Tag(0x3004, 0x000C)

    ds.ImagePositionPatient = origin
    ds.ImageOrientationPatient = direction
    ds.PixelSpacing = spacing[0:2]
    ds.SliceThickness = spacing[2]
    ds.Columns = size[0]
    ds.Rows = size[1]
    ds.GridFrameOffsetVector = [ov * spacing[2] for ov in range(size[2])]
    ds.DoseGridScaling = scaling
    ds.PixelData = (array // scaling).astype(np.uint32).tobytes()
    return ds


# -------------------------------- example --------------------------------------------------- #

if __name__ == "__main__":
    import pydicom
    import SimpleITK as sitk
    dose_sitk_image = sitk.ReadImage("example/data/DF_001/RD_003/CT_001_dose.dcm")
    ds = pydicom.dcmread("example/data/DF_001/RD_003/CT_001.dcm")

    dose_ds = generate_base_dataset()
    dose_ds = cp_information_from_ds(dose_ds, ds)
    dose_ds = add_dose_grid_to_ds(dose_ds, dose_sitk_image)
    dose_ds.save_as("example/data/DF_001/RD_003/dose.dcm")
